#!/usr/bin/env python
import logging
import os
import sys
from datetime import datetime
from os.path import abspath, dirname
from zoneinfo import ZoneInfo

import django
from django.conf import settings
from django.test.runner import DiscoverRunner
from edc_test_utils import DefaultTestSettings

utc = ZoneInfo("UTC")

app_name = "edc_appointment"
base_dir = dirname(abspath(__file__))

DEFAULT_SETTINGS = DefaultTestSettings(
    calling_file=__file__,
    BASE_DIR=base_dir,
    APP_NAME=app_name,
    ETC_DIR=os.path.join(base_dir, app_name, "tests", "etc"),
    EDC_NAVBAR_AUTODISCOVER=False,
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=datetime(2016, 10, 2, 0, 0, 0, tzinfo=utc),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=datetime(2023, 10, 2, 0, 0, 0, tzinfo=utc),
    EDC_AUTH_SKIP_SITE_AUTHS=True,
    EDC_AUTH_SKIP_AUTH_UPDATER=True,
    SUBJECT_SCREENING_MODEL="edc_appointment_app.subjectscreening",
    SUBJECT_CONSENT_MODEL="edc_appointment_app.subjectconsent",
    SUBJECT_VISIT_MODEL="edc_appointment_app.subjectvisit",
    SUBJECT_VISIT_MISSED_MODEL="edc_appointment_app.subjectvisitmissed",
    SUBJECT_REQUISITION_MODEL="edc_appointment_app.subjectrequisition",
    SUBJECT_APP_LABEL="edc_appointment_app",
    INSTALLED_APPS=[
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.staticfiles",
        "django.contrib.sites",
        "django_crypto_fields.apps.AppConfig",
        "django_revision.apps.AppConfig",
        "multisite",
        "edc_auth.apps.AppConfig",
        "edc_action_item.apps.AppConfig",
        "edc_offstudy.apps.AppConfig",
        "edc_consent.apps.AppConfig",
        "edc_crf.apps.AppConfig",
        "edc_dashboard.apps.AppConfig",
        "edc_data_manager.apps.AppConfig",
        "edc_lab.apps.AppConfig",
        "edc_subject_dashboard.apps.AppConfig",
        "edc_device.apps.AppConfig",
        "edc_facility.apps.AppConfig",
        "edc_identifier.apps.AppConfig",
        "edc_list_data.apps.AppConfig",
        "edc_listboard.apps.AppConfig",
        "edc_locator.apps.AppConfig",
        "edc_metadata.apps.AppConfig",
        "edc_model_admin.apps.AppConfig",
        "edc_navbar.apps.AppConfig",
        "edc_protocol.apps.AppConfig",
        "edc_randomization.apps.AppConfig",
        "edc_registration.apps.AppConfig",
        "edc_notification.apps.AppConfig",
        "edc_review_dashboard.apps.AppConfig",
        "edc_sites.apps.AppConfig",
        "edc_timepoint.apps.AppConfig",
        "edc_visit_schedule.apps.AppConfig",
        "edc_visit_tracking.apps.AppConfig",
        "edc_appointment.apps.AppConfig",
        "edc_appointment_app.apps.AppConfig",
    ],
    add_dashboard_middleware=True,
    use_test_urls=True,
).settings


def main():
    if not settings.configured:
        settings.configure(**DEFAULT_SETTINGS)
    django.setup()
    tags = [t.split("=")[1] for t in sys.argv if t.startswith("--tag")]
    failfast = True if [t for t in sys.argv if t == "--failfast"] else False
    failures = DiscoverRunner(failfast=failfast, tags=tags).run_tests([f"{app_name}.tests"])
    sys.exit(failures)


if __name__ == "__main__":
    logging.basicConfig()
    main()



import datetime as dt
from zoneinfo import ZoneInfo

import time_machine
from django.test import TestCase
from edc_consent import site_consents
from edc_facility.import_holidays import import_holidays
from edc_metadata.utils import get_crf_metadata_model_cls
from edc_protocol import Protocol
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED

from edc_appointment.appointment_status_updater import AppointmentStatusUpdater
from edc_appointment.constants import IN_PROGRESS_APPT, INCOMPLETE_APPT, NEW_APPT
from edc_appointment.models import Appointment
from edc_appointment_app.consents import v1_consent
from edc_appointment_app.models import SubjectVisit
from edc_appointment_app.visit_schedule import visit_schedule1, visit_schedule2

from ..helper import Helper

utc_tz = ZoneInfo("UTC")


@time_machine.travel(dt.datetime(2019, 6, 11, 8, 00, tzinfo=utc_tz))
class TestAppointmentStatus(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        self.subject_identifier = "12345"
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule1)
        site_visit_schedules.register(visit_schedule=visit_schedule2)
        site_consents.registry = {}
        site_consents.register(v1_consent)
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            now=Protocol().study_open_datetime,
        )

    def test_appointment_status(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.filter(subject_identifier=self.subject_identifier)
        self.assertEqual(appointments.count(), 4)

        SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )

        self.assertGreater(get_crf_metadata_model_cls().objects.all().count(), 0)
        appointment = appointments[0]
        appointment.refresh_from_db()
        self.assertEqual(appointment.appt_status, IN_PROGRESS_APPT)
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()
        appointment.refresh_from_db()
        self.assertEqual(appointment.appt_status, INCOMPLETE_APPT)
        appointment.appt_status = IN_PROGRESS_APPT
        appointment.save()
        appointment.refresh_from_db()
        self.assertEqual(appointment.appt_status, IN_PROGRESS_APPT)

    def test_appointment_status2(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.filter(subject_identifier=self.subject_identifier)
        self.assertEqual(appointments.count(), 4)

        # complete baseline appt/visit
        appointment_baseline = appointments[0]
        appointment_baseline.appt_status = IN_PROGRESS_APPT
        appointment_baseline.save()
        appointment_baseline.refresh_from_db()

        SubjectVisit.objects.create(
            appointment=appointment_baseline,
            report_datetime=appointment_baseline.appt_datetime,
            reason=SCHEDULED,
        )
        self.assertGreater(get_crf_metadata_model_cls().objects.all().count(), 0)

        appointment = appointments[1]
        appointment.refresh_from_db()
        self.assertEqual(appointment.appt_status, NEW_APPT)
        appointment.appt_status = IN_PROGRESS_APPT
        appointment.save()
        appointment.refresh_from_db()
        appointment_baseline.refresh_from_db()
        self.assertEqual(appointment_baseline.appt_status, INCOMPLETE_APPT)
        self.assertEqual(appointment.appt_status, IN_PROGRESS_APPT)

    def test_appt_status_updater_init(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.filter(subject_identifier=self.subject_identifier)
        self.assertEqual(appointments.count(), 4)

        # complete baseline appt/visit
        appointment_baseline = appointments[0]
        AppointmentStatusUpdater(appointment=appointment_baseline)

    def test_appt_status_updater_init2(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.filter(subject_identifier=self.subject_identifier)
        appointment_baseline = appointments[0]
        appointment_1 = appointments[1]
        appointment_2 = appointments[2]
        appointment_3 = appointments[3]

        self.assertEqual(appointments.count(), 4)

        # complete baseline appt/visit
        appointment_baseline = appointments[0]
        SubjectVisit.objects.create(
            appointment=appointment_baseline,
            report_datetime=appointment_baseline.appt_datetime,
            reason=SCHEDULED,
        )
        appointment_baseline.refresh_from_db()
        self.assertEqual(appointment_baseline.appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointment_1.appt_status, NEW_APPT)
        self.assertEqual(appointment_2.appt_status, NEW_APPT)
        self.assertEqual(appointment_3.appt_status, NEW_APPT)

        # change to incomplete
        appointment_baseline.appt_status = INCOMPLETE_APPT
        appointment_baseline.save()
        appointment_baseline.refresh_from_db()

        self.assertEqual(appointment_baseline.appt_status, INCOMPLETE_APPT)
        self.assertEqual(appointment_1.appt_status, NEW_APPT)
        self.assertEqual(appointment_2.appt_status, NEW_APPT)
        self.assertEqual(appointment_3.appt_status, NEW_APPT)

        # use appointment status updater to change to in_progress
        AppointmentStatusUpdater(appointment=appointment_baseline, change_to_in_progress=True)
        appointment_baseline.refresh_from_db()
        self.assertEqual(appointment_baseline.appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointment_1.appt_status, NEW_APPT)
        self.assertEqual(appointment_2.appt_status, NEW_APPT)
        self.assertEqual(appointment_3.appt_status, NEW_APPT)

    def test_appt_status_updater_appt_1(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.filter(subject_identifier=self.subject_identifier)
        appointment_baseline = appointments[0]
        appointment_1 = appointments[1]
        appointment_2 = appointments[2]
        appointment_3 = appointments[3]

        self.assertEqual(appointments.count(), 4)
        # complete baseline appt/visit
        appointment_baseline = appointments[0]
        SubjectVisit.objects.create(
            appointment=appointment_baseline,
            report_datetime=appointment_baseline.appt_datetime,
            reason=SCHEDULED,
        )
        appointment_baseline.refresh_from_db()
        self.assertEqual(appointment_baseline.appt_status, IN_PROGRESS_APPT)

        # change to incomplete
        appointment_baseline.appt_status = INCOMPLETE_APPT
        appointment_baseline.save()
        appointment_baseline.refresh_from_db()

        SubjectVisit.objects.create(
            appointment=appointment_1,
            report_datetime=appointment_1.appt_datetime,
            reason=SCHEDULED,
        )

        appointment_baseline.refresh_from_db()
        appointment_1.refresh_from_db()
        appointment_2.refresh_from_db()
        appointment_3.refresh_from_db()

        self.assertEqual(appointment_baseline.appt_status, INCOMPLETE_APPT)
        self.assertEqual(appointment_1.appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointment_2.appt_status, NEW_APPT)
        self.assertEqual(appointment_3.appt_status, NEW_APPT)

        appointment_2.appt_status = IN_PROGRESS_APPT
        appointment_2.save()

        appointment_baseline.refresh_from_db()
        appointment_1.refresh_from_db()
        appointment_2.refresh_from_db()
        appointment_3.refresh_from_db()

        self.assertEqual(appointment_baseline.appt_status, INCOMPLETE_APPT)
        self.assertEqual(appointment_1.appt_status, INCOMPLETE_APPT)
        self.assertEqual(appointment_2.appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointment_3.appt_status, NEW_APPT)

    def test_appt_status_updater_appt_1_save_base(self):
        """Using save base and update_fields skips
        AppointmentStatusUpdater in the signal
        """
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.filter(subject_identifier=self.subject_identifier)
        appointment_baseline = appointments[0]
        appointment_1 = appointments[1]
        appointment_2 = appointments[2]
        appointment_3 = appointments[3]

        self.assertEqual(appointments.count(), 4)

        # change
        appointment_baseline.appt_status = IN_PROGRESS_APPT
        appointment_baseline.save_base(update_fields=["appt_status"])
        self.assertEqual(appointment_baseline.appt_status, IN_PROGRESS_APPT)

        # check
        appointment_baseline.refresh_from_db()
        appointment_1.refresh_from_db()
        appointment_2.refresh_from_db()
        appointment_3.refresh_from_db()
        self.assertEqual(appointment_baseline.appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointment_1.appt_status, NEW_APPT)
        self.assertEqual(appointment_2.appt_status, NEW_APPT)
        self.assertEqual(appointment_3.appt_status, NEW_APPT)

        # change
        appointment_1.appt_status = IN_PROGRESS_APPT
        appointment_1.save_base(update_fields=["appt_status"])

        # check
        appointment_baseline.refresh_from_db()
        appointment_1.refresh_from_db()
        appointment_2.refresh_from_db()
        appointment_3.refresh_from_db()
        self.assertEqual(appointment_baseline.appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointment_1.appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointment_2.appt_status, NEW_APPT)
        self.assertEqual(appointment_3.appt_status, NEW_APPT)

    def test_appt_status_updater_appt_1_save_base2(self):
        """Using save base and update_fields skips
        AppointmentStatusUpdater in the signal
        """
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.filter(subject_identifier=self.subject_identifier)
        appointment_baseline = appointments[0]
        appointment_1 = appointments[1]
        appointment_2 = appointments[2]
        appointment_3 = appointments[3]

        self.assertEqual(appointments.count(), 4)

        # change
        appointment_baseline.appt_status = IN_PROGRESS_APPT
        appointment_baseline.save_base()
        SubjectVisit.objects.create(
            appointment=appointment_baseline,
            report_datetime=appointment_baseline.appt_datetime,
            reason=SCHEDULED,
        )
        self.assertEqual(appointment_baseline.appt_status, IN_PROGRESS_APPT)

        # check
        appointment_baseline.refresh_from_db()
        appointment_1.refresh_from_db()
        appointment_2.refresh_from_db()
        appointment_3.refresh_from_db()
        self.assertEqual(appointment_baseline.appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointment_1.appt_status, NEW_APPT)
        self.assertEqual(appointment_2.appt_status, NEW_APPT)
        self.assertEqual(appointment_3.appt_status, NEW_APPT)

        # change
        appointment_1.appt_status = IN_PROGRESS_APPT
        appointment_1.save_base()

        # check
        appointment_baseline.refresh_from_db()
        appointment_1.refresh_from_db()
        appointment_2.refresh_from_db()
        appointment_3.refresh_from_db()
        self.assertEqual(appointment_baseline.appt_status, INCOMPLETE_APPT)
        self.assertEqual(appointment_1.appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointment_2.appt_status, NEW_APPT)
        self.assertEqual(appointment_3.appt_status, NEW_APPT)



from django.apps import apps as django_apps
from django.contrib import admin
from django.core.exceptions import ObjectDoesNotExist
from django_audit_fields import audit_fieldset_tuple
from edc_crf.admin import crf_status_fieldset_tuple

from ..choices import APPT_DATE_INFO_SOURCES


class NextAppointmentCrfModelAdminMixin:
    fieldsets = (
        (None, {"fields": ("subject_visit", "report_datetime")}),
        (
            "Integrated Clinic",
            {
                "fields": (
                    "info_source",
                    "health_facility",
                    "appt_date",
                    "allow_create_interim",
                    "visitschedule",
                )
            },
        ),
        crf_status_fieldset_tuple,
        audit_fieldset_tuple,
    )

    radio_fields = {
        "crf_status": admin.VERTICAL,
        "info_source": admin.VERTICAL,
    }

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "visitschedule":
            visit_schedule_model_cls = django_apps.get_model(
                "edc_visit_schedule.visitschedule"
            )
            try:
                related_visit = self.related_visit(request)
            except ObjectDoesNotExist:
                kwargs["queryset"] = visit_schedule_model_cls.objects.none()
            else:
                kwargs["queryset"] = visit_schedule_model_cls.objects.filter(
                    visit_schedule_name=related_visit.visit_schedule_name,
                    schedule_name=related_visit.schedule_name,
                    active=True,
                )
        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    def formfield_for_choice_field(self, db_field, request, **kwargs):
        if db_field.name == "info_source":
            kwargs["choices"] = APPT_DATE_INFO_SOURCES
        return super().formfield_for_choice_field(db_field, request, **kwargs)



import datetime as dt
from unittest.mock import patch
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import relativedelta
from django.test import TestCase, override_settings
from edc_facility.import_holidays import import_holidays
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import MISSED_VISIT, SCHEDULED, UNSCHEDULED
from tqdm import tqdm

from edc_appointment.constants import (
    COMPLETE_APPT,
    INCOMPLETE_APPT,
    MISSED_APPT,
    ONTIME_APPT,
)
from edc_appointment.creators import (
    UnscheduledAppointmentCreator,
    UnscheduledAppointmentError,
)
from edc_appointment.exceptions import AppointmentWindowError
from edc_appointment.forms import AppointmentForm
from edc_appointment.models import Appointment
from edc_appointment_app.models import SubjectVisit
from edc_appointment_app.visit_schedule import visit_schedule3
from edc_appointment_app.visit_schedule.visit_schedule4 import visit_schedule4

from ...utils import (
    AppointmentDateWindowPeriodGapError,
    get_appointment_by_datetime,
    get_window_gap_days,
)
from ..helper import Helper

utc = ZoneInfo("UTC")


@time_machine.travel(dt.datetime(2019, 6, 11, 8, 00, tzinfo=utc))
class TestAppointmentWindowPeriod(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        self.subject_identifier = "12345"
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule3)
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            now=get_utcnow() - relativedelta(years=2),  # Protocol().study_open_datetime,
        )

    @staticmethod
    def create_unscheduled(appointment):
        return UnscheduledAppointmentCreator(
            subject_identifier=appointment.subject_identifier,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
            suggested_visit_code_sequence=appointment.visit_code_sequence + 1,
            facility=appointment.facility,
        ).appointment

    @staticmethod
    def get_form(appointment, appt_datetime):
        return AppointmentForm(
            data={
                "appt_datetime": appt_datetime,
                "appt_timing": ONTIME_APPT,
                "subject_identifier": appointment.subject_identifier,
                "timepoint_status": appointment.timepoint_status,
                "timepoint": appointment.timepoint,
                "timepoint_datetime": appointment.timepoint_datetime,
                "appt_close_datetime": appointment.timepoint_datetime,
                "facility_name": appointment.facility_name,
                "appt_type": appointment.appt_type,
                "appt_status": appointment.appt_status,
                "appt_reason": appointment.appt_reason,
                "document_status": appointment.document_status,
            },
            instance=appointment,
        )

    @staticmethod
    def update_appt_as_incomplete(appointment):
        appointment.refresh_from_db()
        SubjectVisit.objects.create(
            appointment=appointment,
            report_datetime=appointment.appt_datetime,
            reason=SCHEDULED if appointment.visit_code_sequence == 0 else UNSCHEDULED,
            schedule_name=appointment.schedule_name,
            visit_code=appointment.visit_code,
            visit_code_sequence=appointment.visit_code_sequence,
        )
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()
        appointment.refresh_from_db()

    def create_1030_and_1060(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule3",
            schedule_name="three_monthly_schedule",
        )

        appointment_1000 = Appointment.objects.get(
            subject_identifier=self.subject_identifier, visit_code="1000"
        )
        SubjectVisit.objects.create(
            appointment=appointment_1000,
            report_datetime=appointment_1000.appt_datetime,
            reason=SCHEDULED,
        )
        appointment_1000.appt_status = INCOMPLETE_APPT
        appointment_1000.save()
        appointment_1000.refresh_from_db()

        appointment_1030 = Appointment.objects.get(
            subject_identifier=self.subject_identifier, visit_code="1030"
        )
        SubjectVisit.objects.create(
            appointment=appointment_1030,
            report_datetime=appointment_1030.appt_datetime,
            reason=SCHEDULED,
        )
        appointment_1030.appt_status = INCOMPLETE_APPT
        appointment_1030.save()
        appointment_1030.refresh_from_db()
        return (
            appointment_1030,
            Appointment.objects.get(
                subject_identifier=self.subject_identifier, visit_code="1060"
            ),
        )

    def test_appointments_window_period(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule3",
            schedule_name="three_monthly_schedule",
        )
        appointments = Appointment.objects.filter(subject_identifier=self.subject_identifier)
        self.assertEqual(appointments.count(), 5)

        appointment_1030 = Appointment.objects.get(
            subject_identifier=self.subject_identifier, visit_code="1030"
        )
        appointment_1060 = Appointment.objects.get(
            subject_identifier=self.subject_identifier, visit_code="1060"
        )
        appointment_1030.appt_datetime = appointment_1060.appt_datetime - relativedelta(days=1)
        self.assertRaises(AppointmentWindowError, appointment_1030.save)

    @patch("edc_appointment.form_validators.utils.url_names")
    def test_appointments_window_period_in_form(self, mock_urlnames):
        mock_urlnames.return_value = {"subject_dashboard_url": "subject_dashboard_url"}
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule3",
            schedule_name="three_monthly_schedule",
        )
        appointment_1030 = Appointment.objects.get(
            subject_identifier=self.subject_identifier, visit_code="1030"
        )
        appointment_1060 = Appointment.objects.get(
            subject_identifier=self.subject_identifier, visit_code="1060"
        )
        form = AppointmentForm(
            data={"appt_datetime": appointment_1060.appt_datetime},
            instance=appointment_1030,
        )
        form.is_valid()
        # conflicts on non-unique appointment date
        self.assertIn("This appointment conflicts", form._errors.get("appt_datetime")[0])
        form = AppointmentForm(
            data={"appt_datetime": appointment_1060.appt_datetime - relativedelta(days=1)},
            instance=appointment_1030,
        )
        form.is_valid()
        # outside of window period
        self.assertIn("appt_datetime", form._errors)
        self.assertIn(
            "Invalid. Expected a date/time between", form._errors.get("appt_datetime")[0]
        )

    def test_appointments_window_period_boundary_before_next_lower(self):
        appointment_1030, appointment_1060 = self.create_1030_and_1060()

        delta = appointment_1060.appt_datetime - appointment_1030.appt_datetime

        appt_datetime_before = (
            appointment_1030.appt_datetime
            + delta
            - appointment_1060.visit_from_schedule.rlower
            - relativedelta(days=1)
        )
        unscheduled_appointment = self.create_unscheduled(appointment_1030)
        form = self.get_form(unscheduled_appointment, appt_datetime_before)
        form.is_valid()
        self.assertNotIn("appt_datetime", form._errors)
        form.save()
        self.update_appt_as_incomplete(unscheduled_appointment)

    def test_appointments_window_period_boundary_on_next_lower(self):
        appointment_1030, appointment_1060 = self.create_1030_and_1060()

        delta = appointment_1060.appt_datetime - appointment_1030.appt_datetime

        appt_datetime_on = (
            appointment_1030.appt_datetime
            + delta
            - appointment_1060.visit_from_schedule.rlower
        )
        unscheduled_appointment = self.create_unscheduled(appointment_1030)
        form = self.get_form(unscheduled_appointment, appt_datetime_on)
        form.is_valid()
        self.assertIn("appt_datetime", form._errors)

    def test_appointments_window_period_boundary_after_next_lower(self):
        appointment_1030, appointment_1060 = self.create_1030_and_1060()

        delta = appointment_1060.appt_datetime - appointment_1030.appt_datetime

        appt_datetime_after = (
            appointment_1030.appt_datetime
            + delta
            - appointment_1060.visit_from_schedule.rlower
            + relativedelta(days=1)
        )
        unscheduled_appointment = self.create_unscheduled(appointment_1030)
        form = self.get_form(unscheduled_appointment, appt_datetime_after)
        form.is_valid()
        self.assertIn("appt_datetime", form._errors)

    @override_settings(EDC_APPOINTMENT_CHECK_APPT_STATUS=False)
    def test_appointments_window_period_allows_between_completed_appointments(self):
        appointment_1030, appointment_1060 = self.create_1030_and_1060()

        # assert both appointments are complete or incomplete
        appointment_1030.appt_status = INCOMPLETE_APPT
        appointment_1030.save()
        appointment_1060.appt_status = INCOMPLETE_APPT
        appointment_1060.save()
        self.assertIn(appointment_1030.appt_status, [INCOMPLETE_APPT, COMPLETE_APPT])
        self.assertIn(appointment_1060.appt_status, [INCOMPLETE_APPT, COMPLETE_APPT])

        # get appt_date one day before 1060
        appt_datetime_after_1030 = appointment_1060.appt_datetime - relativedelta(days=1)
        # create unscheduled off of 1030
        unscheduled_appointment = self.create_unscheduled(appointment_1030)

        # set appt_date to  one day before 1060
        form = self.get_form(unscheduled_appointment, appt_datetime_after_1030)
        form.is_valid()
        self.assertNotIn("appt_datetime", form._errors)

    def test_appointments_window_period_does_not_allow_between_new_appointments(self):
        appointment_1030, appointment_1060 = self.create_1030_and_1060()

        # assert only appointment_1030 is complete / incomplete
        appointment_1030.appt_status = INCOMPLETE_APPT
        appointment_1030.save()
        self.assertIn(appointment_1030.appt_status, [INCOMPLETE_APPT, COMPLETE_APPT])
        self.assertNotIn(
            appointment_1060.appt_status, [INCOMPLETE_APPT, COMPLETE_APPT, MISSED_APPT]
        )

        # get appt_date one day before 1060
        appt_datetime_after_1030 = appointment_1060.appt_datetime - relativedelta(days=1)
        # create unscheduled off of 1030
        unscheduled_appointment = self.create_unscheduled(appointment_1030)

        # set appt_date to  one day before 1060
        form = self.get_form(unscheduled_appointment, appt_datetime_after_1030)
        form.is_valid()
        self.assertIn("appt_datetime", form._errors)

    @override_settings(EDC_APPOINTMENT_CHECK_APPT_STATUS=False)
    def test_appointments_window_period_does_not_allow_missed_unscheduled(self):
        """Assert does not allow an unscheduled appointment if the
        scheduled appt is missed (in this case 1030 is missed)
        """
        appointment_1030, appointment_1060 = self.create_1030_and_1060()

        appointment_1030.appt_timing = MISSED_APPT
        appointment_1030.save()
        appointment_1030.refresh_from_db()

        subject_visit = SubjectVisit.objects.get(appointment=appointment_1030)
        subject_visit.reason = MISSED_VISIT
        subject_visit.save()
        subject_visit.refresh_from_db()

        appointment_1030.appt_status = INCOMPLETE_APPT
        appointment_1030.save()
        appointment_1030.refresh_from_db()

        self.assertIn(subject_visit.reason, [MISSED_VISIT])
        self.assertIn(appointment_1030.appt_timing, [MISSED_APPT])
        self.assertIn(appointment_1030.appt_status, [INCOMPLETE_APPT])

        # get appt_date one day before 1060
        appt_datetime_after_1030 = appointment_1060.appt_datetime - relativedelta(days=1)
        # create unscheduled off of 1030
        unscheduled_appointment = self.create_unscheduled(appointment_1030)

        # set appt_date to  one day before 1060
        form = self.get_form(unscheduled_appointment, appt_datetime_after_1030)
        form.is_valid()
        self.assertIn("appt_datetime", form._errors)

    @override_settings(EDC_APPOINTMENT_CHECK_APPT_STATUS=False)
    def test_appointments_window_period_unscheduled(self):
        appointment_1030, appointment_1060 = self.create_1030_and_1060()

        appointment_1060_lower_appt_datetime = (
            appointment_1060.appt_datetime - appointment_1060.visit.rlower
        )

        # assert only appointment_1030 is complete / incomplete
        appointment_1030.appt_status = INCOMPLETE_APPT
        appointment_1030.save()
        # create unscheduled off of 1030 until you hit lower bound of
        # 1060 window
        unscheduled_appointment = self.create_unscheduled(appointment_1030)
        unscheduled_appointment.appt_status = INCOMPLETE_APPT
        unscheduled_appointment.save()
        total = (
            appointment_1060_lower_appt_datetime - unscheduled_appointment.appt_datetime
        ).days
        for _ in tqdm(range(1, total + 5), total=total):
            unscheduled_appointment = self.create_unscheduled(unscheduled_appointment)
            unscheduled_appointment.appt_status = INCOMPLETE_APPT
            unscheduled_appointment.save()
            if (
                unscheduled_appointment.appt_datetime
                == appointment_1060_lower_appt_datetime - relativedelta(days=1)
            ):
                break

        # attempt to hit lower bound of 1060 window raises exception
        self.assertRaises(
            UnscheduledAppointmentError,
            self.create_unscheduled,
            unscheduled_appointment,
        ),

        self.assertEqual(
            (
                appointment_1060_lower_appt_datetime - unscheduled_appointment.appt_datetime
            ).days,
            1,
        )

        # now use the form to try to create the next unscheduled visit on the
        # date of the lower bound of 1060.
        form = self.get_form(
            unscheduled_appointment,
            appointment_1060_lower_appt_datetime,
        )
        form.is_valid()
        # form.save(commit=True)
        self.assertIn("appt_datetime", form._errors)
        self.assertIn(
            "Falls outside of the window period", form._errors.get("appt_datetime")[0]
        )

    def test_match_appt_date_to_visit_code(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule3",
            schedule_name="three_monthly_schedule",
        )
        appointments = Appointment.objects.filter(
            subject_identifier=self.subject_identifier
        ).order_by("appt_datetime")
        self.assertEqual(appointments.count(), 5)
        appointment_1000 = appointments[0]

        suggested_appt_datetime = appointment_1000.appt_datetime + relativedelta(months=3)
        appointment = get_appointment_by_datetime(
            suggested_appt_datetime,
            appointment_1000.subject_identifier,
            appointment_1000.visit_schedule_name,
            appointment_1000.schedule_name,
        )
        self.assertIsNotNone(appointment)
        self.assertEqual(appointment.visit_code, "1030")

        suggested_appt_datetime = appointment_1000.appt_datetime + relativedelta(months=6)
        appointment = get_appointment_by_datetime(
            suggested_appt_datetime,
            appointment_1000.subject_identifier,
            appointment_1000.visit_schedule_name,
            appointment_1000.schedule_name,
        )
        self.assertIsNotNone(appointment)
        self.assertEqual(appointment.visit_code, "1060")

    def test_match_appt_date_to_visit_code2(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule3",
            schedule_name="three_monthly_schedule",
        )
        appointments = Appointment.objects.filter(
            subject_identifier=self.subject_identifier
        ).order_by("appt_datetime")
        self.assertEqual(appointments.count(), 5)
        appointment_1000 = appointments[0]
        appointment_1030 = appointments[1]
        appointment_1120 = appointments[4]

        suggested_appt_datetime = (
            appointment_1030.appt_datetime + appointment_1030.visit.rupper
        )
        appointment = get_appointment_by_datetime(
            suggested_appt_datetime,
            appointment_1000.subject_identifier,
            appointment_1000.visit_schedule_name,
            appointment_1000.schedule_name,
        )
        self.assertIsNotNone(appointment)
        self.assertEqual(appointment.visit_code, "1030")

        suggested_appt_datetime = (
            appointment_1030.appt_datetime
            + appointment_1030.visit.rupper
            + relativedelta(days=1)
        )
        with self.assertRaises(AppointmentDateWindowPeriodGapError) as cm:
            get_appointment_by_datetime(
                suggested_appt_datetime,
                appointment_1000.subject_identifier,
                appointment_1000.visit_schedule_name,
                appointment_1000.schedule_name,
            )
        self.assertIn(
            "Date falls in a `window period gap` between 1030 and 1060", str(cm.exception)
        )

        suggested_appt_datetime = (
            appointment_1030.appt_datetime
            + appointment_1030.visit.rupper
            + relativedelta(months=2)
        )
        appointment = get_appointment_by_datetime(
            suggested_appt_datetime,
            appointment_1000.subject_identifier,
            appointment_1000.visit_schedule_name,
            appointment_1000.schedule_name,
        )

        self.assertIsNotNone(appointment)
        self.assertEqual(appointment.visit_code, "1060")

        suggested_appt_datetime = (
            appointment_1120.appt_datetime
            + appointment_1120.visit.rupper
            + relativedelta(days=1)
        )
        appointment = get_appointment_by_datetime(
            suggested_appt_datetime,
            appointment_1000.subject_identifier,
            appointment_1000.visit_schedule_name,
            appointment_1000.schedule_name,
        )
        self.assertIsNone(appointment)


@time_machine.travel(dt.datetime(2019, 7, 11, 8, 00, tzinfo=utc))
class TestAppointmentWindowPeriod2(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        self.subject_identifier = "12345"
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule4)
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            now=get_utcnow() - relativedelta(years=2),
        )

    @override_settings(EDC_VISIT_SCHEDULE_DEFAULT_MAX_VISIT_GAP_ALLOWED=0)
    def test_suggested_date_in_window_period_gap_raises(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule4",
            schedule_name="three_monthly_schedule",
        )
        appointments = Appointment.objects.filter(
            subject_identifier=self.subject_identifier
        ).order_by("appt_datetime")
        self.assertEqual(appointments.count(), 5)
        appointment_1000 = appointments[0]
        appointment_1060 = appointments[2]

        subject_identifier = appointment_1000.subject_identifier
        visit_schedule_name = appointment_1000.visit_schedule_name
        schedule_name = appointment_1000.schedule_name

        suggested_appt_datetime = (
            appointment_1060.appt_datetime
            - appointment_1060.visit.rlower
            - relativedelta(days=1)
        )
        with self.assertRaises(AppointmentDateWindowPeriodGapError) as cm:
            get_appointment_by_datetime(
                suggested_appt_datetime,
                subject_identifier,
                visit_schedule_name,
                schedule_name,
            )

        self.assertIn(
            "Date falls in a `window period gap` between 1030 and 1060", str(cm.exception)
        )

    @override_settings(EDC_VISIT_SCHEDULE_DEFAULT_MAX_VISIT_GAP_ALLOWED=7)
    def test_match_window_period_gap_adjusted(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule4",
            schedule_name="three_monthly_schedule",
        )
        appointments = Appointment.objects.filter(
            subject_identifier=self.subject_identifier
        ).order_by("appt_datetime")
        self.assertEqual(appointments.count(), 5)
        appointment_1000 = appointments[0]
        appointment_1060 = appointments[2]

        subject_identifier = appointment_1000.subject_identifier
        visit_schedule_name = appointment_1000.visit_schedule_name
        schedule_name = appointment_1000.schedule_name

        for days, in_adjusted_window in [
            (6, True),
            (7, True),
            (8, False),
            (9, False),
        ]:
            with self.subTest(days=days, in_adjusted_window=in_adjusted_window):
                suggested_appt_datetime = (
                    appointment_1060.appt_datetime
                    - appointment_1060.visit.rlower
                    - relativedelta(days=days)
                )
                try:
                    appointment = get_appointment_by_datetime(
                        suggested_appt_datetime,
                        subject_identifier,
                        visit_schedule_name,
                        schedule_name,
                        raise_if_in_gap=False,
                    )
                except AppointmentDateWindowPeriodGapError as e:
                    self.fail(
                        f"AppointmentDateWindowPeriodGapError unexpectedly raised. Got {e}"
                    )
                if not in_adjusted_window:
                    self.assertIsNone(appointment)
                else:
                    self.assertEqual("1060", appointment.visit_code)

    def test_past_last_visit(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule4",
            schedule_name="three_monthly_schedule",
        )
        appointments = Appointment.objects.filter(
            subject_identifier=self.subject_identifier
        ).order_by("appt_datetime")
        self.assertEqual(appointments.count(), 5)
        appointment_1000 = appointments[0]
        appointment_1120 = appointments[4]

        subject_identifier = appointment_1000.subject_identifier
        visit_schedule_name = appointment_1000.visit_schedule_name
        schedule_name = appointment_1000.schedule_name

        suggested_appt_datetime = (
            appointment_1120.appt_datetime
            + appointment_1120.visit.rupper
            + relativedelta(days=1)
        )
        try:
            appointment = get_appointment_by_datetime(
                suggested_appt_datetime,
                subject_identifier,
                visit_schedule_name,
                schedule_name,
                raise_if_in_gap=False,
            )
        except AppointmentDateWindowPeriodGapError as e:
            self.fail(f"AppointmentDateWindowPeriodGapError unexpectedly raised. Got {e}")
        self.assertIsNone(appointment)

    def test_window_gap_days(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule4",
            schedule_name="three_monthly_schedule",
        )
        appointments = Appointment.objects.filter(
            subject_identifier=self.subject_identifier
        ).order_by("appt_datetime")
        appointment_1030 = appointments[1]
        gap_days = get_window_gap_days(appointment_1030)
        self.assertEqual(gap_days, 62)



CANCELLED_APPT = "cancelled"
COMPLETE_APPT = "done"
INCOMPLETE_APPT = "incomplete"
IN_PROGRESS_APPT = "in_progress"
LAST_MONTH = "last_month"
MISSED_APPT = "missed"
NEW_APPT = "new"
NEXT_MONTH = "next_month"
ONTIME_APPT = "ontime"
SCHEDULED_APPT = "scheduled"
UNSCHEDULED_APPT = "unscheduled"
SKIPPED_APPT = "skipped"



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any

from dateutil.relativedelta import relativedelta
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.db import IntegrityError
from edc_utils import formatted_datetime, to_utc
from edc_utils.date import to_local
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_schedule.utils import get_lower_datetime

from ..constants import (
    COMPLETE_APPT,
    IN_PROGRESS_APPT,
    INCOMPLETE_APPT,
    NEW_APPT,
    UNSCHEDULED_APPT,
)
from ..exceptions import AppointmentPermissionsRequired, AppointmentWindowError
from .appointment_creator import AppointmentCreator, CreateAppointmentError

if TYPE_CHECKING:
    from edc_facility import Facility

    from ..models import Appointment


class UnscheduledAppointmentError(Exception):
    pass


class UnscheduledAppointmentNotAllowed(Exception):
    pass


class InvalidVisitCodeSequencesError(Exception):
    pass


class InvalidTimepointError(Exception):
    pass


class InvalidParentAppointmentStatusError(Exception):
    pass


class InvalidParentAppointmentMissingVisitError(Exception):
    pass


class AppointmentInProgressError(Exception):
    pass


class UnscheduledAppointmentCreator:

    """Attempts to create a new unscheduled appointment where the
    visit code sequence == the given `visit_code_sequence` or raises.
    """

    appointment_creator_cls = AppointmentCreator

    def __init__(
        self,
        subject_identifier: str = None,
        visit_schedule_name: str = None,
        schedule_name: str = None,
        visit_code: str = None,
        suggested_visit_code_sequence: int = None,
        suggested_appt_datetime: datetime | None = None,
        facility: Facility = None,
        request: Any | None = None,
    ):
        self._parent_appointment = None
        self._calling_appointment = None
        self.appointment = None
        self._suggested_appt_datetime = suggested_appt_datetime
        self.subject_identifier = subject_identifier
        self.visit_schedule_name = visit_schedule_name
        self.schedule_name = schedule_name
        self.visit_code = visit_code
        self.visit_code_sequence = suggested_visit_code_sequence
        if self.visit_code_sequence is None:
            raise InvalidVisitCodeSequencesError("visit code sequence cannot be None")
        if self.visit_code_sequence < 1:
            raise InvalidVisitCodeSequencesError(
                "suggested visit code sequence cannot be less than 1"
            )
        self.facility = facility
        self.visit_schedule = site_visit_schedules.get_visit_schedule(visit_schedule_name)
        self.schedule = self.visit_schedule.schedules.get(schedule_name)
        self.appointment_model_cls = self.schedule.appointment_model_cls
        self.has_perm_or_raise(request)
        self.create_or_raise()

    def create_or_raise(self) -> None:
        visit = self.visit_schedule.schedules.get(self.schedule_name).visits.get(
            self.visit_code
        )
        if not visit:
            raise UnscheduledAppointmentError(
                "Invalid visit. Got None using {"
                f"visit_schedule_name='{self.visit_schedule_name}',"
                f"schedule_name='{self.schedule_name}',"
                f"visit_code='{self.visit_code}'" + "}"
            )
        elif not visit.allow_unscheduled:
            raise UnscheduledAppointmentNotAllowed(
                f"Not allowed. Visit {self.visit_code} is not configured for "
                "unscheduled appointments."
            )
        elif (
            self.parent_appointment
            and self.parent_appointment.next
            and to_utc(self.suggested_appt_datetime)
            >= get_lower_datetime(self.parent_appointment.next)
        ):
            dt = formatted_datetime(to_local(self.suggested_appt_datetime))
            next_dt = formatted_datetime(
                to_local(get_lower_datetime(self.parent_appointment.next))
            )
            raise UnscheduledAppointmentError(
                "Appointment date exceeds window period. Next appointment is "
                f"{self.parent_appointment.next.visit_code} and lower window starts "
                f"on {next_dt}. Got {dt}. Perhaps catch this in the form."
            )
        else:
            # do not allow new unscheduled if any appointments are IN_PROGRESS
            try:
                obj = self.appointment_model_cls.objects.get(
                    subject_identifier=self.subject_identifier,
                    visit_schedule_name=self.visit_schedule_name,
                    schedule_name=self.schedule_name,
                    appt_status=IN_PROGRESS_APPT,
                )
            except MultipleObjectsReturned as e:
                raise UnscheduledAppointmentError(e)
            except ObjectDoesNotExist:
                pass
            else:
                raise AppointmentInProgressError(
                    f"Not allowed. Appointment {obj.visit_code}."
                    f"{obj.visit_code_sequence} is in progress."
                )
            try:
                appointment_creator = self.appointment_creator_cls(
                    subject_identifier=self.subject_identifier,
                    visit_schedule_name=self.visit_schedule_name,
                    schedule_name=self.schedule_name,
                    visit=visit,
                    suggested_datetime=self.suggested_appt_datetime,
                    timepoint=self.parent_appointment.timepoint,
                    timepoint_datetime=self.parent_appointment.timepoint_datetime,
                    visit_code_sequence=self.parent_appointment.next_visit_code_sequence,
                    facility=self.facility,
                    appt_status=NEW_APPT,
                    appt_reason=UNSCHEDULED_APPT,
                    ignore_window_period=self.ignore_window_period,
                )
            except CreateAppointmentError:
                raise
            except IntegrityError as e:
                raise CreateAppointmentError(e)
            except AppointmentWindowError as e:
                msg = str(e).replace("Perhaps catch this in the form", "")
                raise UnscheduledAppointmentError(
                    f"Unable to create unscheduled appointment. {msg}"
                )
            self.appointment = appointment_creator.appointment

    def has_perm_or_raise(self, request) -> None:
        if request and not request.user.has_perm(
            f"{self.appointment_model_cls._meta.app_label}."
            f"add_{self.appointment_model_cls._meta.model_name}"
        ):
            raise AppointmentPermissionsRequired(
                "You do not have permission to create an appointment"
            )

    @property
    def ignore_window_period(self: Any) -> bool:
        value = False
        if (
            self.calling_appointment
            and self.calling_appointment.next
            and self.calling_appointment.next.appt_status in [INCOMPLETE_APPT, COMPLETE_APPT]
            and self.suggested_appt_datetime < self.calling_appointment.next.appt_datetime
        ):
            value = True
        return value

    @property
    def suggested_appt_datetime(self):
        if not self._suggested_appt_datetime:
            self._suggested_appt_datetime = (
                self.calling_appointment.appt_datetime + relativedelta(days=1)
            )
        return self._suggested_appt_datetime

    @property
    def calling_appointment(self) -> Appointment:
        """Returns the appointment which preceeds the to-be-created
        unscheduled appt.

        It must exist.

        Note: subtracting 1 from `visit_code_sequence` implies the
        visit_code_sequence cannot be less than 1.

        if visit_code_sequence = 1, calling_appointment and parent_appointment
        will be the same instance.
        """
        if not self._calling_appointment:
            opts = dict(
                subject_identifier=self.subject_identifier,
                visit_schedule_name=self.visit_schedule_name,
                schedule_name=self.schedule_name,
                visit_code=self.visit_code,
                visit_code_sequence=self.visit_code_sequence - 1,
                timepoint=self.parent_appointment.timepoint,
            )
            self._calling_appointment = self.appointment_model_cls.objects.get(**opts)
        return self._calling_appointment

    @property
    def parent_appointment(self) -> Appointment:
        """Returns the parent or 'scheduled' appointment
        (visit_code_sequence=0).
        """
        if not self._parent_appointment:
            options = dict(
                subject_identifier=self.subject_identifier,
                visit_schedule_name=self.visit_schedule_name,
                schedule_name=self.schedule_name,
                visit_code=self.visit_code,
                visit_code_sequence=0,
            )
            self._parent_appointment = self.appointment_model_cls.objects.get(**options)
            if not self._parent_appointment.related_visit:
                raise InvalidParentAppointmentMissingVisitError(
                    f"Unable to create unscheduled appointment. An unscheduled "
                    f"appointment cannot be created if the parent appointment "
                    f"visit form has not been completed. "
                    f"Got appointment '{self.visit_code}.0'."
                )
            else:
                if self._parent_appointment.appt_status not in [
                    COMPLETE_APPT,
                    INCOMPLETE_APPT,
                ]:
                    raise InvalidParentAppointmentStatusError(
                        f"Unable to create unscheduled appointment. An unscheduled "
                        f"appointment cannot be created if the parent appointment "
                        f"is 'new' or 'in progress'. Got appointment "
                        f"'{self.visit_code}' is "
                        f"{self._parent_appointment.get_appt_status_display().lower()}."
                    )

        return self._parent_appointment



# Generated by Django 2.0rc1 on 2017-11-24 09:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_appointment", "0009_auto_20171119_1032")]

    operations = [
        migrations.AlterField(
            model_name="appointment",
            name="facility_name",
            field=models.CharField(
                help_text="set by model that creates appointments, e.g. Enrollment",
                max_length=25,
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="facility_name",
            field=models.CharField(
                help_text="set by model that creates appointments, e.g. Enrollment",
                max_length=25,
            ),
        ),
    ]



from .visit_schedule1 import visit_schedule1
from .visit_schedule2 import visit_schedule2
from .visit_schedule3 import visit_schedule3
from .visit_schedule5 import visit_schedule5
from .visit_schedule6 import visit_schedule6



from __future__ import annotations

from typing import TYPE_CHECKING

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from edc_constants.constants import INCOMPLETE
from edc_visit_tracking.constants import MISSED_VISIT

from ..appointment_reason_updater import AppointmentReasonUpdater
from ..constants import IN_PROGRESS_APPT, MISSED_APPT

if TYPE_CHECKING:
    from ..models import Appointment


class MissedAppointmentModelMixin(models.Model):
    def create_missed_visit_from_appointment(self: Appointment):
        if self.appt_timing == MISSED_APPT:
            try:
                subject_visit = self.related_visit_model_cls().objects.get(
                    appointment__id=self.id
                )
            except ObjectDoesNotExist:
                self.related_visit_model_cls().objects.create_missed_from_appointment(
                    appointment=self,
                )
            else:
                subject_visit.reason = MISSED_VISIT
                subject_visit.document_status = INCOMPLETE
                subject_visit.save_base(update_fields=["reason", "document_status"])

    def update_subject_visit_reason_or_raise(self: Appointment):
        """Trys to update the subject_visit.reason field, if it
        exists, when appt_timing changes, or raises.
        """
        if (
            self.id
            and self.related_visit
            and self.appt_status == IN_PROGRESS_APPT
            and self.appt_timing
            and self.appt_reason
        ):
            AppointmentReasonUpdater(
                appointment=self,
                appt_timing=self.appt_timing,
                appt_reason=self.appt_reason,
                commit=True,
            )

    class Meta:
        abstract = True



from datetime import datetime
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import relativedelta
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase, override_settings
from edc_consent import site_consents
from edc_facility import import_holidays
from edc_visit_schedule.apps import populate_visit_schedule
from edc_visit_schedule.models import VisitSchedule
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED

from edc_appointment.constants import (
    COMPLETE_APPT,
    IN_PROGRESS_APPT,
    INCOMPLETE_APPT,
    NEW_APPT,
    SKIPPED_APPT,
)
from edc_appointment.models import Appointment
from edc_appointment.skip_appointments import (
    SkipAppointmentsFieldError,
    SkipAppointmentsValueError,
)
from edc_appointment.tests.helper import Helper
from edc_appointment.tests.test_case_mixins import AppointmentTestCaseMixin
from edc_appointment.utils import get_allow_skipped_appt_using
from edc_appointment_app.consents import v1_consent
from edc_appointment_app.forms import CrfThreeForm
from edc_appointment_app.models import (
    CrfFive,
    CrfFour,
    CrfOne,
    CrfThree,
    CrfTwo,
    SubjectVisit,
)
from edc_appointment_app.visit_schedule import (
    visit_schedule1,
    visit_schedule2,
    visit_schedule5,
)

utc = ZoneInfo("UTC")


@time_machine.travel(datetime(2019, 6, 11, 8, 00, tzinfo=utc))
class TestSkippedAppt(AppointmentTestCaseMixin, TestCase):
    helper_cls = Helper

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        self.subject_identifier = "12345"
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule1)
        site_visit_schedules.register(visit_schedule=visit_schedule2)
        site_visit_schedules.register(visit_schedule=visit_schedule5)
        site_consents.registry = {}
        site_consents.register(v1_consent)
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            now=datetime(2017, 6, 5, 8, 0, 0, tzinfo=utc),
        )
        populate_visit_schedule()

    @override_settings(
        EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING={
            "edc_appointment_app.crfone": ("next_appt_date", "next_visit_code")
        }
    )
    def test_skip_appointments_using_crf_date(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        appointments[0].appt_status = IN_PROGRESS_APPT
        appointments[0].save()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        appointments[0].refresh_from_db()
        self.assertEqual(appointments[0].appt_status, IN_PROGRESS_APPT)

        self.assertRaises(
            SkipAppointmentsValueError,
            CrfOne.objects.create,
            subject_visit=subject_visit,
            next_appt_date=subject_visit.report_datetime + relativedelta(weeks=3),
        )
        CrfOne.objects.all().delete()

        CrfOne.objects.create(
            subject_visit=subject_visit,
            next_appt_date=subject_visit.report_datetime + relativedelta(weeks=3),
            next_visit_code=appointments[3].visit_code,
        )

        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")

        self.assertEqual(appointments[0].appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointments[1].appt_status, SKIPPED_APPT)

    @override_settings(
        EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING={
            "edc_appointment_app.crfone": ("next_appt_date", "next_visit_code"),
            "edc_appointment_app.crftwo": ("report_datetime", "next_visit_code"),
            "edc_appointment_app.crfthree": ("report_datetime", "f1"),
            "edc_appointment_app.crffour": ("report_datetime", "next_visit_code"),
        }
    )
    def test_settings_improperly_configured(self):
        self.assertRaises(ImproperlyConfigured, get_allow_skipped_appt_using)

    @override_settings(
        EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING={
            "edc_appointment_app.crftwo": ("report_datetime", "next_visit_code_blah"),
        }
    )
    def test_skip_appointments_using_bad_settings_for_crf(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        appointments[0].appt_status = IN_PROGRESS_APPT
        appointments[0].save()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        appointments[0].refresh_from_db()
        self.assertEqual(appointments[0].appt_status, IN_PROGRESS_APPT)

        # next_visit_code_blah is not a valid field on Crf
        self.assertRaises(
            SkipAppointmentsFieldError,
            CrfTwo.objects.create,
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
        )

    @override_settings(
        EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING={
            "edc_appointment_app.crftwo": ("report_datetime_blah", "f1"),
        }
    )
    def test_skip_appointments_using_bad_settings_for_crf2(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        appointments[0].appt_status = IN_PROGRESS_APPT
        appointments[0].save()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        appointments[0].refresh_from_db()
        self.assertEqual(appointments[0].appt_status, IN_PROGRESS_APPT)

        # report_datetime_blah is not a valid field on Crf
        self.assertRaises(
            SkipAppointmentsFieldError,
            CrfTwo.objects.create,
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            f1=appointments[3].visit_code,
        )

    @override_settings(
        EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING={
            "edc_appointment_app.crfthree": ("report_datetime", "f1"),
        }
    )
    def test_skip_appointments_using_crf_ok(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        appointments[0].appt_status = IN_PROGRESS_APPT
        appointments[0].save()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        appointments[0].refresh_from_db()
        self.assertEqual(appointments[0].appt_status, IN_PROGRESS_APPT)
        try:
            CrfThree.objects.create(
                subject_visit=subject_visit,
                report_datetime=appointments[3].appt_datetime,
                f1=appointments[3].visit_code,
            )
        except SkipAppointmentsValueError as e:
            self.fail(f"SkipAppointmentsValueError unexpectedly raised. Got {e}")
        except SkipAppointmentsFieldError as e:
            self.fail(f"SkipAppointmentsFieldError unexpectedly raised. Got {e}")

    @override_settings(
        EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING={
            "edc_appointment_app.crfthree": ("report_datetime", "f1"),
        }
    )
    def test_skip_multiple_appointments_using_good_crf(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        appointments[0].appt_status = IN_PROGRESS_APPT
        appointments[0].save()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        appointments[0].refresh_from_db()

        CrfThree.objects.create(
            subject_visit=subject_visit,
            report_datetime=appointments[3].appt_datetime,
            f1=appointments[3].visit_code,
        )
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[1].appt_status, SKIPPED_APPT)
        self.assertEqual(appointments[2].appt_status, SKIPPED_APPT)
        self.assertEqual(appointments[3].appt_status, NEW_APPT)

    @override_settings(
        EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING={
            "edc_appointment_app.crfthree": ("appt_date", "f1"),
        }
    )
    def test_skip_multiple_appointments_using_last_crf(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        appointments[0].appt_status = IN_PROGRESS_APPT
        appointments[0].save()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        CrfThree.objects.create(
            subject_visit=subject_visit,
            appt_date=appointments[2].appt_datetime.date(),
            f1=appointments[2].visit_code,
        )

        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointments[1].appt_status, SKIPPED_APPT)
        self.assertEqual(appointments[2].appt_status, NEW_APPT)
        self.assertEqual(appointments[3].appt_status, NEW_APPT)

        appointments[1].appt_status = IN_PROGRESS_APPT
        appointments[1].save()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[1],
            report_datetime=appointments[1].appt_datetime,
            reason=SCHEDULED,
        )
        CrfThree.objects.create(
            subject_visit=subject_visit,
            f1=appointments[3].visit_code,
            appt_date=appointments[3].appt_datetime.date(),
        )

        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_status, INCOMPLETE_APPT)
        self.assertEqual(appointments[1].appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointments[2].appt_status, SKIPPED_APPT)
        self.assertEqual(appointments[3].appt_status, NEW_APPT)

    @override_settings(
        EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING={
            "edc_appointment_app.crftwo": ("report_datetime", "visitschedule"),
        }
    )
    def test_visit_code_as_visit_schedule_fk_ok(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        appointments[0].appt_status = IN_PROGRESS_APPT
        appointments[0].save()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        self.assertRaises(
            SkipAppointmentsValueError,
            CrfTwo.objects.create,
            subject_visit=subject_visit,
            report_datetime=appointments[1].appt_datetime,
            visitschedule=VisitSchedule.objects.get(visit_code=appointments[3].visit_code),
        )

        self.assertRaises(
            SkipAppointmentsValueError,
            CrfTwo.objects.create,
            subject_visit=subject_visit,
            report_datetime=appointments[3].appt_datetime,
            visitschedule=VisitSchedule.objects.get(visit_code=appointments[1].visit_code),
        )

    @override_settings(
        EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING={
            "edc_appointment_app.crfthree": ("appt_date", "f1"),
        }
    )
    def test_last_crf_with_absurd_date_relative_to_visit_code(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        appointments[0].appt_status = IN_PROGRESS_APPT
        appointments[0].save()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        self.assertRaises(
            SkipAppointmentsValueError,
            CrfThree.objects.create,
            subject_visit=subject_visit,
            appt_date=appointments[1].appt_datetime.date(),
            f1=appointments[3].visit_code,
        )

    @override_settings(
        EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING={
            "edc_appointment_app.crfthree": ("appt_date", "f1"),
        }
    )
    def test_delete(self):
        self.helper.consent_and_put_on_schedule()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        appointments[0].appt_status = IN_PROGRESS_APPT
        appointments[0].save()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        crf_three_1000 = CrfThree.objects.create(
            subject_visit=subject_visit,
            appt_date=appointments[2].appt_datetime.date(),
            f1=appointments[2].visit_code,
        )
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        appointments[1].appt_status = IN_PROGRESS_APPT
        appointments[1].save()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[1],
            report_datetime=appointments[1].appt_datetime,
            reason=SCHEDULED,
        )
        crf_three_2000 = CrfThree.objects.create(
            subject_visit=subject_visit,
            f1=appointments[3].visit_code,
            appt_date=appointments[3].appt_datetime.date(),
        )
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_status, INCOMPLETE_APPT)
        self.assertEqual(appointments[1].appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointments[2].appt_status, SKIPPED_APPT)
        self.assertEqual(appointments[3].appt_status, NEW_APPT)

        crf_three_2000.delete()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_status, INCOMPLETE_APPT)
        self.assertEqual(appointments[1].appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointments[2].appt_status, NEW_APPT)
        self.assertEqual(appointments[3].appt_status, NEW_APPT)

        crf_three_1000.delete()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_status, INCOMPLETE_APPT)
        self.assertEqual(appointments[1].appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointments[2].appt_status, NEW_APPT)
        self.assertEqual(appointments[3].appt_status, NEW_APPT)

    @override_settings(
        EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING={
            "edc_appointment_app.crfthree": ("appt_date", "f1"),
        }
    )
    def test_skip2(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule5", schedule_name="monthly_schedule"
        )
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        appointments[0].appt_status = IN_PROGRESS_APPT
        appointments[0].save()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )

        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointments[1].appt_status, NEW_APPT)

        for model_cls in [CrfOne, CrfTwo, CrfFour, CrfFive]:
            model_cls.objects.create(
                subject_visit=subject_visit,
                report_datetime=appointments[0].appt_datetime,
                f1="blah",
            )
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointments[1].appt_status, NEW_APPT)
        CrfThree.objects.create(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            appt_date=appointments[2].appt_datetime.date(),
            f1=appointments[2].visit_code,
        )
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointments[1].appt_status, SKIPPED_APPT)
        self.assertEqual(appointments[2].appt_status, NEW_APPT)
        self.assertEqual(appointments[3].appt_status, NEW_APPT)
        self.assertEqual(appointments[4].appt_status, NEW_APPT)
        self.assertEqual(appointments[5].appt_status, NEW_APPT)
        self.assertEqual(appointments[6].appt_status, NEW_APPT)

        appointments[2].appt_status = IN_PROGRESS_APPT
        appointments[2].save()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[2],
            report_datetime=appointments[2].appt_datetime,
            reason=SCHEDULED,
        )

        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_status, COMPLETE_APPT)
        self.assertEqual(appointments[1].appt_status, SKIPPED_APPT)
        self.assertEqual(appointments[2].appt_status, IN_PROGRESS_APPT)

        for model_cls in [CrfOne, CrfTwo, CrfFour, CrfFive]:
            model_cls.objects.create(
                subject_visit=subject_visit,
                report_datetime=appointments[2].appt_datetime,
                f1="blah",
            )

        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_status, COMPLETE_APPT)
        self.assertEqual(appointments[1].appt_status, SKIPPED_APPT)
        self.assertEqual(appointments[2].appt_status, IN_PROGRESS_APPT)

        CrfThree.objects.create(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            appt_date=appointments[3].appt_datetime.date(),
            f1=appointments[3].visit_code,
        )

        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_status, COMPLETE_APPT)
        self.assertEqual(appointments[1].appt_status, SKIPPED_APPT)
        self.assertEqual(appointments[2].appt_status, IN_PROGRESS_APPT)

        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        appointments[2].appt_status = COMPLETE_APPT
        appointments[2].save()
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_status, COMPLETE_APPT)
        self.assertEqual(appointments[1].appt_status, SKIPPED_APPT)
        self.assertEqual(appointments[2].appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointments[3].appt_status, NEW_APPT)
        self.assertEqual(appointments[4].appt_status, NEW_APPT)
        self.assertEqual(appointments[5].appt_status, NEW_APPT)
        self.assertEqual(appointments[6].appt_status, NEW_APPT)

        appointments[3].appt_status = IN_PROGRESS_APPT
        appointments[3].save()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[3],
            report_datetime=appointments[3].appt_datetime,
            reason=SCHEDULED,
        )
        CrfOne.objects.create(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            f1="blah",
        )
        CrfThree.objects.create(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            appt_date=appointments[6].appt_datetime.date(),
            f1=appointments[6].visit_code,
        )

        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_status, COMPLETE_APPT)
        self.assertEqual(appointments[1].appt_status, SKIPPED_APPT)
        self.assertEqual(appointments[2].appt_status, COMPLETE_APPT)
        self.assertEqual(appointments[3].appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointments[4].appt_status, SKIPPED_APPT)
        self.assertEqual(appointments[5].appt_status, SKIPPED_APPT)
        self.assertEqual(appointments[6].appt_status, NEW_APPT)

        appointments[0].appt_status = IN_PROGRESS_APPT
        appointments[0].save()
        subject_visit = SubjectVisit.objects.get(appointment=appointments[0])
        subject_visit.save()
        CrfThree.objects.create(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            appt_date=appointments[2].appt_datetime.date(),
            f1=appointments[2].visit_code,
        )
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[0].appt_status, IN_PROGRESS_APPT)
        self.assertEqual(appointments[1].appt_status, SKIPPED_APPT)
        self.assertEqual(appointments[2].appt_status, COMPLETE_APPT)
        self.assertEqual(appointments[3].appt_status, INCOMPLETE_APPT)
        self.assertEqual(appointments[4].appt_status, SKIPPED_APPT)
        self.assertEqual(appointments[5].appt_status, SKIPPED_APPT)
        self.assertEqual(appointments[6].appt_status, NEW_APPT)

    # @time_machine.travel(datetime(2021, 10, 6, 8, 00, tzinfo=utc))
    @override_settings(
        EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING={
            "edc_appointment_app.crfthree": ("appt_date", "f1"),
        }
    )
    def test_when_next_appointment_in_window(self):
        self.helper.consent_and_put_on_schedule(
            visit_schedule_name="visit_schedule5", schedule_name="monthly_schedule"
        )
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        appointments[0].appt_status = IN_PROGRESS_APPT
        appointments[0].save()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )

        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        for model_cls in [CrfOne, CrfTwo, CrfFour, CrfFive]:
            model_cls.objects.create(
                subject_visit=subject_visit,
                report_datetime=appointments[0].appt_datetime,
                f1="blah",
            )
        form = CrfThreeForm(
            data=dict(
                subject_visit=subject_visit,
                report_datetime=appointments[0].appt_datetime,
                appt_date=appointments[0].appt_datetime + relativedelta(days=3),
                f1=appointments[1].visit_code,
            )
        )
        form.is_valid()
        self.assertEqual({}, form._errors)
        form.save(commit=True)

        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(
            (appointments[0].appt_datetime + relativedelta(days=3)).date(),
            appointments[1].appt_datetime.date(),
        )

        traveller = time_machine.travel(appointments[0].appt_datetime + relativedelta(days=3))
        traveller.start()
        appointments[1].appt_status = IN_PROGRESS_APPT
        appointments[1].save()
        subject_visit = SubjectVisit.objects.create(
            appointment=appointments[1],
            report_datetime=appointments[1].appt_datetime,
            reason=SCHEDULED,
        )
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        for model_cls in [CrfOne, CrfTwo, CrfFour, CrfFive]:
            model_cls.objects.create(
                subject_visit=subject_visit,
                report_datetime=appointments[1].appt_datetime,
                f1="blah",
            )
        form = CrfThreeForm(
            data=dict(
                subject_visit=subject_visit,
                report_datetime=appointments[1].appt_datetime,
                appt_date=appointments[1].appt_datetime + relativedelta(days=5),
                f1=appointments[1].visit_code,
                allow_create_interim=True,
            )
        )
        form.is_valid()
        self.assertEqual({}, form._errors)
        form.save(commit=True)

        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        self.assertEqual(appointments[2].visit_code, "1010")
        self.assertEqual(appointments[2].visit_code_sequence, 1)

        self.assertEqual(
            (appointments[1].appt_datetime + relativedelta(days=5)).date(),
            appointments[2].appt_datetime.date(),
        )



from django import forms
from edc_crf.crf_form_validator import CrfFormValidator
from edc_crf.modelform_mixins import CrfModelFormMixin

from edc_appointment.form_validator_mixins import NextAppointmentCrfFormValidatorMixin
from edc_appointment.modelform_mixins import NextAppointmentCrfModelFormMixin

from .models import CrfThree, NextAppointmentCrf


class NextAppointmentCrfFormValidator(NextAppointmentCrfFormValidatorMixin, CrfFormValidator):
    def clean(self):
        self.validate_date_is_on_clinic_day()
        super().clean()


class CrfThreeForm(NextAppointmentCrfModelFormMixin, CrfModelFormMixin, forms.ModelForm):
    form_validator_cls = NextAppointmentCrfFormValidator

    appt_date_fld = "appt_date"
    visit_code_fld = "f1"

    def validate_against_consent(self) -> None:
        pass

    class Meta:
        model = CrfThree
        fields = "__all__"
        labels = {"appt_date": "Next scheduled appointment date"}


class NextAppointmentCrfForm(
    NextAppointmentCrfModelFormMixin, CrfModelFormMixin, forms.ModelForm
):
    form_validator_cls = NextAppointmentCrfFormValidator

    def validate_against_consent(self) -> None:
        pass

    class Meta:
        model = NextAppointmentCrf
        fields = "__all__"



from __future__ import annotations

from django import forms
from edc_form_validators import FormValidatorMixin
from edc_model_form.mixins import BaseModelFormMixin
from edc_offstudy.modelform_mixins import OffstudyNonCrfModelFormMixin
from edc_sites.forms import SiteModelFormMixin
from edc_visit_schedule.modelform_mixins import VisitScheduleNonCrfModelFormMixin

from ..form_validators import AppointmentFormValidator
from ..models import Appointment
from ..utils import get_appointment_form_meta_options

appt_reason_fld = Appointment._meta.get_field("appt_reason")
appt_type_fld = Appointment._meta.get_field("appt_type")


class AppointmentForm(
    SiteModelFormMixin,
    VisitScheduleNonCrfModelFormMixin,
    OffstudyNonCrfModelFormMixin,
    BaseModelFormMixin,
    FormValidatorMixin,
    forms.ModelForm,
):
    """Note, the appointment is only changed, never added,
    through this form.
    """

    form_validator_cls = AppointmentFormValidator
    report_datetime_field_attr: str = "appt_datetime"

    class Meta:
        model = Appointment
        fields = "__all__"
        labels = get_appointment_form_meta_options()["labels"]
        help_texts = get_appointment_form_meta_options()["help_texts"]



from datetime import datetime
from decimal import Context
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import FR, MO, SA, SU, TH, TU, WE, relativedelta
from django.core.exceptions import ImproperlyConfigured, ObjectDoesNotExist
from django.db import transaction
from django.db.models.deletion import ProtectedError
from django.test import TestCase, override_settings
from edc_consent import site_consents
from edc_constants.constants import INCOMPLETE
from edc_facility.import_holidays import import_holidays
from edc_protocol import Protocol
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import MISSED_VISIT, SCHEDULED, UNSCHEDULED
from edc_visit_tracking.exceptions import RelatedVisitReasonError
from edc_visit_tracking.utils import get_related_visit_model_cls

from edc_appointment.constants import (
    IN_PROGRESS_APPT,
    INCOMPLETE_APPT,
    MISSED_APPT,
    ONTIME_APPT,
    SCHEDULED_APPT,
    UNSCHEDULED_APPT,
)
from edc_appointment.exceptions import (
    AppointmentBaselineError,
    AppointmentDatetimeError,
)
from edc_appointment.managers import AppointmentDeleteError
from edc_appointment.utils import get_appointment_model_cls, get_appt_reason_choices
from edc_appointment_app.consents import v1_consent
from edc_appointment_app.models import OnScheduleOne, OnScheduleTwo, SubjectConsent
from edc_appointment_app.visit_schedule import visit_schedule1, visit_schedule2

from ..helper import Helper

utc_tz = ZoneInfo("UTC")


@time_machine.travel(datetime(2019, 6, 11, 8, 00, tzinfo=utc_tz))
class TestAppointment(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        self.subject_identifier = "12345"
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule1)
        site_visit_schedules.register(visit_schedule=visit_schedule2)
        site_consents.registry = {}
        site_consents.register(v1_consent)
        self.helper = self.helper_cls(
            subject_identifier=self.subject_identifier,
            now=Protocol().study_open_datetime,
        )

    def test_appointments_creation(self):
        """Assert appointment triggering method creates appointments."""
        self.helper.consent_and_put_on_schedule()
        appointments = get_appointment_model_cls().objects.filter(
            subject_identifier=self.subject_identifier
        )
        self.assertEqual(appointments.count(), 4)

    def test_appointments_creation2(self):
        """Asserts first appointment correctly selected if
        both visit_schedule_name and schedule_name provided.
        """
        self.helper.consent_and_put_on_schedule()
        OnScheduleTwo.objects.create(
            subject_identifier=self.subject_identifier, onschedule_datetime=get_utcnow()
        )
        self.assertEqual(get_appointment_model_cls().objects.all().count(), 8)

    def test_deletes_appointments(self):
        """Asserts manager method can delete appointments."""
        self.helper.consent_and_put_on_schedule()
        appointments = get_appointment_model_cls().objects.filter(
            subject_identifier=self.subject_identifier
        )
        self.assertEqual(appointments.count(), 4)

        get_related_visit_model_cls().objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )

        visit_schedule = site_visit_schedules.get_visit_schedule(
            visit_schedule_name=appointments[0].visit_schedule_name
        )
        schedule = visit_schedule.schedules.get(appointments[0].schedule_name)

        # this calls the manager method "delete_for_subject_after_date"
        schedule.offschedule_model_cls.objects.create(
            subject_identifier=self.subject_identifier,
            offschedule_datetime=appointments[0].appt_datetime,
        )

        self.assertEqual(
            get_appointment_model_cls()
            .objects.filter(subject_identifier=self.subject_identifier)
            .count(),
            1,
        )

    def test_deletes_appointments_with_unscheduled(self):
        """Asserts manager method can delete appointments."""
        self.helper.consent_and_put_on_schedule()
        appointments = get_appointment_model_cls().objects.filter(
            subject_identifier=self.subject_identifier
        )
        self.assertEqual(appointments.count(), 4)

        appointment = appointments[0]

        get_related_visit_model_cls().objects.create(
            appointment=appointment,
            report_datetime=appointment.appt_datetime,
            reason=SCHEDULED,
        )

        appointment.refresh_from_db()
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()
        appointment.refresh_from_db()

        self.helper.add_unscheduled_appointment(appointment)
        self.assertEqual(appointments.count(), 5)

        visit_schedule = site_visit_schedules.get_visit_schedule(
            visit_schedule_name=appointment.visit_schedule_name
        )
        schedule = visit_schedule.schedules.get(appointment.schedule_name)

        # this calls the manager method "delete_for_subject_after_date"
        schedule.take_off_schedule(
            subject_identifier=self.subject_identifier,
            offschedule_datetime=appointment.appt_datetime,
        )

        self.assertEqual(
            get_appointment_model_cls()
            .objects.filter(subject_identifier=self.subject_identifier)
            .count(),
            1,
        )

    def test_delete_single_appointment(self):
        self.helper.consent_and_put_on_schedule()
        appointments = get_appointment_model_cls().objects.filter(
            subject_identifier=self.subject_identifier
        )
        self.assertEqual(appointments.count(), 4)

        get_related_visit_model_cls().objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )

        appointment = appointments[0]
        appointment.refresh_from_db()
        self.assertEqual(appointment.appt_status, IN_PROGRESS_APPT)
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()
        appointment.refresh_from_db()
        self.assertEqual(appointment.appt_status, INCOMPLETE_APPT)

        # raised by Django
        with transaction.atomic():
            self.assertRaises(ProtectedError, appointment.delete)

        # raised in signal pre_delete
        with transaction.atomic():
            self.assertRaises(AppointmentDeleteError, appointments[1].delete)

        self.helper.add_unscheduled_appointment(appointment)
        self.assertEqual(appointments.count(), 5)

        unscheduled_appointment = get_appointment_model_cls().objects.get(
            visit_code="1000", visit_code_sequence=1
        )

        unscheduled_appointment.delete()
        self.assertEqual(appointments.count(), 4)

    def test_appointments_dates_mo(self):
        """Test appointment datetimes are chronological."""
        for index in range(0, 7):
            SubjectConsent.objects.create(
                subject_identifier=f"{self.subject_identifier}-{index}",
                consent_datetime=get_utcnow(),
            )
        for day in [MO, TU, WE, TH, FR, SA, SU]:
            subject_consent = SubjectConsent.objects.all()[day.weekday]
            OnScheduleOne.objects.create(
                subject_identifier=subject_consent.subject_identifier,
                onschedule_datetime=(
                    subject_consent.consent_datetime
                    + relativedelta(weeks=2)
                    + relativedelta(weekday=day(-1))
                ),
            )
            appt_datetimes = [
                obj.appt_datetime
                for obj in get_appointment_model_cls()
                .objects.filter(subject_identifier=subject_consent.subject_identifier)
                .order_by("appt_datetime")
            ]
            last = None
            self.assertGreater(len(appt_datetimes), 0)
            for appt_datetime in appt_datetimes:
                if not last:
                    last = appt_datetime
                else:
                    self.assertGreater(appt_datetime, last)
                    last = appt_datetime

    def test_attempt_to_change(self):
        for index in range(0, 7):
            SubjectConsent.objects.create(
                subject_identifier=f"{self.subject_identifier}-{index}",
                consent_datetime=get_utcnow(),
            )
        for day in [MO, TU, WE, TH, FR, SA, SU]:
            subject_consent = SubjectConsent.objects.all()[day.weekday]
            OnScheduleOne.objects.create(
                subject_identifier=subject_consent.subject_identifier,
                onschedule_datetime=(
                    subject_consent.consent_datetime
                    + relativedelta(weeks=2)
                    + relativedelta(weekday=day(-1))
                ),
            )
        subject_identifier = subject_consent.subject_identifier
        self.assertEqual(
            get_appointment_model_cls()
            .objects.filter(subject_identifier=subject_identifier)
            .count(),
            4,
        )

        # create two unscheduled appts after first
        appointment0 = (
            get_appointment_model_cls()
            .objects.filter(subject_identifier=subject_identifier)
            .order_by("appt_datetime")[0]
        )
        get_related_visit_model_cls().objects.create(
            appointment=appointment0,
            report_datetime=appointment0.appt_datetime,
            reason=SCHEDULED,
        )
        appointment0.appt_status = INCOMPLETE_APPT
        appointment0.save()

        appointment0_1 = self.helper.add_unscheduled_appointment(appointment0)
        get_related_visit_model_cls().objects.create(
            appointment=appointment0_1,
            report_datetime=appointment0.appt_datetime,
            reason=UNSCHEDULED,
        )
        appointment0_1.appt_status = INCOMPLETE_APPT
        appointment0_1.save()

        appointment0_2 = self.helper.add_unscheduled_appointment(appointment0_1)

        appointment0_1.appt_datetime = appointment0_2.appt_datetime + relativedelta(days=1)

        self.assertRaises(AppointmentDatetimeError, appointment0_1.save)

    def test_timepoint(self):
        """Assert timepoints are saved from the schedule correctly
        as Decimals and ordered by appt_datetime.
        """
        context = Context(prec=2)
        self.helper.consent_and_put_on_schedule()
        self.assertEqual(
            [
                obj.timepoint
                for obj in get_appointment_model_cls().objects.all().order_by("appt_datetime")
            ],
            [context.create_decimal(n) for n in range(0, 4)],
        )

    def test_first_appointment_with_visit_schedule_and_schedule(self):
        """Asserts first appointment correctly selected if
        both visit_schedule_name and schedule_name provided.
        """
        self.helper.consent_and_put_on_schedule()
        subject_consent = SubjectConsent.objects.get(
            subject_identifier=self.subject_identifier
        )
        OnScheduleTwo.objects.create(
            subject_identifier=subject_consent.subject_identifier,
            onschedule_datetime=subject_consent.consent_datetime,
        )
        onschedule_one = OnScheduleOne.objects.get(subject_identifier=self.subject_identifier)
        first_appointment = get_appointment_model_cls().objects.first_appointment(
            subject_identifier=onschedule_one.subject_identifier,
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )

        appointment = (
            get_appointment_model_cls()
            .objects.filter(
                subject_identifier=onschedule_one.subject_identifier,
                visit_schedule_name="visit_schedule1",
                schedule_name="schedule1",
            )
            .order_by("appt_datetime")[0]
        )

        self.assertEqual(first_appointment, appointment)

    def test_first_appointment_with_visit_schedule(self):
        """Asserts first appointment correctly selected if just
        visit_schedule_name provided.
        """
        subject_consent = SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier, consent_datetime=get_utcnow()
        )
        OnScheduleTwo.objects.create(
            subject_identifier=self.subject_identifier,
            onschedule_datetime=subject_consent.consent_datetime,
        )
        OnScheduleOne.objects.create(
            subject_identifier=self.subject_identifier,
            onschedule_datetime=(subject_consent.report_datetime + relativedelta(months=1)),
        )

        first_appointment = get_appointment_model_cls().objects.first_appointment(
            subject_identifier=self.subject_identifier,
            visit_schedule_name="visit_schedule2",
        )
        self.assertEqual(first_appointment.schedule_name, "schedule2")

        self.assertEqual(
            get_appointment_model_cls()
            .objects.filter(
                subject_identifier=self.subject_identifier,
                visit_schedule_name="visit_schedule2",
            )
            .order_by("appt_datetime")[0],
            first_appointment,
        )

    def test_first_appointment_with_unscheduled(self):
        """Asserts first appointment correctly selected if
        unscheduled visits have been added.
        """
        self.helper.consent_and_put_on_schedule()
        subject_consent = SubjectConsent.objects.get(
            subject_identifier=self.subject_identifier
        )
        OnScheduleTwo.objects.create(
            subject_identifier=subject_consent.subject_identifier,
            onschedule_datetime=subject_consent.consent_datetime,
        )
        onschedule_one = OnScheduleOne.objects.get(subject_identifier=self.subject_identifier)

        first_appointment = get_appointment_model_cls().objects.first_appointment(
            subject_identifier=onschedule_one.subject_identifier,
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )

        # add unscheduled
        get_related_visit_model_cls().objects.create(
            appointment=first_appointment,
            report_datetime=first_appointment.appt_datetime,
            reason=SCHEDULED,
        )
        first_appointment.appt_status = INCOMPLETE_APPT
        first_appointment.save()
        self.helper.add_unscheduled_appointment(first_appointment)

        appointment = (
            get_appointment_model_cls()
            .objects.filter(
                subject_identifier=onschedule_one.subject_identifier,
                visit_schedule_name="visit_schedule1",
                schedule_name="schedule1",
            )
            .order_by("appt_datetime")[0]
        )
        self.assertEqual(first_appointment, appointment)

    def test_next_appointment(self):
        self.helper.consent_and_put_on_schedule()
        onschedule = OnScheduleOne.objects.get(subject_identifier=self.subject_identifier)
        first_appointment = get_appointment_model_cls().objects.first_appointment(
            subject_identifier=onschedule.subject_identifier,
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        next_appointment = get_appointment_model_cls().objects.next_appointment(
            visit_code=first_appointment.visit_code,
            subject_identifier=onschedule.subject_identifier,
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        self.assertEqual(
            get_appointment_model_cls()
            .objects.filter(subject_identifier=onschedule.subject_identifier)
            .order_by("appt_datetime")[1],
            next_appointment,
        )

        next_appointment = get_appointment_model_cls().objects.next_appointment(
            appointment=first_appointment
        )
        self.assertEqual(
            get_appointment_model_cls()
            .objects.filter(subject_identifier=onschedule.subject_identifier)
            .order_by("appt_datetime")[1],
            next_appointment,
        )

    def test_next_appointment_with_unscheduled(self):
        self.helper.consent_and_put_on_schedule()
        onschedule = OnScheduleOne.objects.get(subject_identifier=self.subject_identifier)
        first_appointment = get_appointment_model_cls().objects.first_appointment(
            subject_identifier=onschedule.subject_identifier,
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )

        # add unscheduled
        get_related_visit_model_cls().objects.create(
            appointment=first_appointment,
            report_datetime=first_appointment.appt_datetime,
            reason=SCHEDULED,
        )
        first_appointment.appt_status = INCOMPLETE_APPT
        first_appointment.save()
        first_appointment.refresh_from_db()
        self.helper.add_unscheduled_appointment(first_appointment)

        next_appointment = get_appointment_model_cls().objects.next_appointment(
            visit_code=first_appointment.visit_code,
            subject_identifier=onschedule.subject_identifier,
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        self.assertEqual(
            get_appointment_model_cls()
            .objects.filter(
                subject_identifier=onschedule.subject_identifier, visit_code_sequence=0
            )
            .order_by("timepoint")[1],
            next_appointment,
        )

        next_appointment = get_appointment_model_cls().objects.next_appointment(
            appointment=first_appointment
        )
        self.assertEqual(
            get_appointment_model_cls()
            .objects.filter(
                subject_identifier=onschedule.subject_identifier, visit_code_sequence=0
            )
            .order_by("timepoint")[1],
            next_appointment,
        )

    def test_next_appointment_after_last_returns_none(self):
        """Assert returns None if next after last appointment."""
        self.helper.consent_and_put_on_schedule()

        last_appointment = get_appointment_model_cls().objects.last_appointment(
            subject_identifier=self.subject_identifier,
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        self.assertEqual(
            get_appointment_model_cls().objects.next_appointment(appointment=last_appointment),
            None,
        )

    def test_next_appointment_after_last_returns_none_with_unscheduled(self):
        """Assert returns None if next after last appointment."""
        self.helper.consent_and_put_on_schedule()

        last_appointment = get_appointment_model_cls().objects.last_appointment(
            subject_identifier=self.subject_identifier,
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )

        # add unscheduled
        for appointment in get_appointment_model_cls().objects.all():
            appointment.appt_status = IN_PROGRESS_APPT
            appointment.save()
            get_related_visit_model_cls().objects.create(
                appointment=appointment,
                report_datetime=appointment.appt_datetime,
                reason=SCHEDULED,
            )
            appointment.appt_status = INCOMPLETE_APPT
            appointment.save()
        self.helper.add_unscheduled_appointment(last_appointment)

        self.assertEqual(
            get_appointment_model_cls().objects.next_appointment(appointment=last_appointment),
            None,
        )

    def test_next_appointment_until_none(self):
        """Assert can walk from first to last appointment."""
        self.helper.consent_and_put_on_schedule()
        appointments = get_appointment_model_cls().objects.filter(
            subject_identifier=self.subject_identifier,
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        first = get_appointment_model_cls().objects.first_appointment(
            appointment=appointments[0]
        )
        appts = [first]
        for appointment in appointments:
            appts.append(
                get_appointment_model_cls().objects.next_appointment(appointment=appointment)
            )
        self.assertIsNotNone(appts[0])
        self.assertEqual(appts[0], first)
        self.assertEqual(appts[-1], None)

    def test_previous_appointment1(self):
        """Assert returns None if relative to first appointment."""
        self.helper.consent_and_put_on_schedule()
        first_appointment = get_appointment_model_cls().objects.first_appointment(
            subject_identifier=self.subject_identifier,
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        self.assertEqual(
            get_appointment_model_cls().objects.previous_appointment(
                appointment=first_appointment
            ),
            None,
        )

    def test_previous_appointment2(self):
        """Assert returns previous appointment."""
        self.helper.consent_and_put_on_schedule()
        first_appointment = get_appointment_model_cls().objects.first_appointment(
            subject_identifier=self.subject_identifier,
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        next_appointment = get_appointment_model_cls().objects.next_appointment(
            appointment=first_appointment
        )
        self.assertEqual(
            get_appointment_model_cls().objects.previous_appointment(
                appointment=next_appointment
            ),
            first_appointment,
        )

    def test_next_and_previous_appointment3(self):
        """Assert accepts appointment or indiviual attrs."""
        self.helper.consent_and_put_on_schedule()
        first_appointment = get_appointment_model_cls().objects.first_appointment(
            subject_identifier=self.subject_identifier,
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        next_appointment = get_appointment_model_cls().objects.next_appointment(
            visit_code=first_appointment.visit_code,
            subject_identifier=self.subject_identifier,
            visit_schedule_name="visit_schedule1",
            schedule_name="schedule1",
        )
        self.assertEqual(
            get_appointment_model_cls()
            .objects.filter(subject_identifier=self.subject_identifier)
            .order_by("appt_datetime")[1],
            next_appointment,
        )
        appointment = get_appointment_model_cls().objects.next_appointment(
            appointment=first_appointment
        )
        self.assertEqual(
            get_appointment_model_cls()
            .objects.filter(subject_identifier=self.subject_identifier)
            .order_by("appt_datetime")[1],
            appointment,
        )

    @override_settings(
        EDC_APPOINTMENT_APPT_REASON_CHOICES=((UNSCHEDULED_APPT, "Unscheduled"),)
    )
    def test_setting_reason_choices_invalid(self):
        self.assertRaises(ImproperlyConfigured, get_appt_reason_choices)

    @override_settings(
        EDC_APPOINTMENT_APPT_REASON_CHOICES=(
            (SCHEDULED_APPT, "Routine / Scheduled"),
            (UNSCHEDULED_APPT, "Unscheduled"),
        )
    )
    def test_setting_reason_choices_valid(self):
        try:
            choices = get_appt_reason_choices()
        except ImproperlyConfigured:
            self.fail("ImproperlyConfigured unexpectedly raised")
        self.assertEqual(
            choices,
            (
                (SCHEDULED_APPT, "Routine / Scheduled"),
                (UNSCHEDULED_APPT, "Unscheduled"),
            ),
        )

    @override_settings(
        EDC_APPOINTMENT_APPT_REASON_CHOICES=(
            (SCHEDULED_APPT, "Routine / Scheduled"),
            (UNSCHEDULED_APPT, "Unscheduled"),
            ("blah", "Blah"),
        )
    )
    def test_setting_reason_choices_valid2(self):
        try:
            choices = get_appt_reason_choices()
        except ImproperlyConfigured:
            self.fail("ImproperlyConfigured unexpectedly raised")
        self.assertEqual(
            choices,
            (
                (SCHEDULED_APPT, "Routine / Scheduled"),
                (UNSCHEDULED_APPT, "Unscheduled"),
                ("blah", "Blah"),
            ),
        )

    def test_raises_appt_timing_missed_at_baseline(self):
        self.helper.consent_and_put_on_schedule()
        appointments = (
            get_appointment_model_cls()
            .objects.filter(subject_identifier=self.subject_identifier)
            .order_by("timepoint")
        )
        self.assertEqual(appointments.count(), 4)
        appointment = appointments.first()
        appointment.appt_status = IN_PROGRESS_APPT
        appointment.appt_timing = MISSED_APPT
        self.assertRaises(AppointmentBaselineError, appointment.save)

    def test_appt_timing(self):
        self.helper.consent_and_put_on_schedule()
        appointments = (
            get_appointment_model_cls()
            .objects.filter(subject_identifier=self.subject_identifier)
            .order_by("timepoint")
        )
        self.assertEqual(appointments.count(), 4)
        # create report for baseline visit
        appointment = appointments.first()
        get_related_visit_model_cls().objects.create(
            appointment=appointment,
            report_datetime=appointment.appt_datetime,
            reason=SCHEDULED,
        )
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save()

        # next appt
        appointment = appointment.next
        appointment.appt_timing = MISSED_APPT
        appointment.save()
        appointment.refresh_from_db()
        self.assertEqual(appointment.appt_timing, MISSED_APPT)

    def test_appointment_creates_subject_visit_if_missed(self):
        self.helper.consent_and_put_on_schedule()
        appointments = get_appointment_model_cls().objects.filter(
            subject_identifier=self.subject_identifier
        )
        self.assertEqual(appointments.count(), 4)
        # create report for baseline visit
        get_related_visit_model_cls().objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        appointment = get_appointment_model_cls().objects.get(id=appointments[1].id)
        appointment.appt_timing = MISSED_APPT
        appointment.save()
        appointment.refresh_from_db()
        try:
            subject_visit = get_related_visit_model_cls().objects.get(
                appointment__visit_code=appointment.visit_code
            )
        except ObjectDoesNotExist:
            self.fail("ObjectDoesNotExist unexpectedly raised")
        self.assertEqual(subject_visit.reason, MISSED_VISIT)
        self.assertIn("auto-created", subject_visit.comments)
        self.assertEqual(subject_visit.document_status, INCOMPLETE)

        # resave does not cause error
        appointment.save()

    def test_raises_if_subject_visit_reason_out_of_sync_with_appt(self):
        self.helper.consent_and_put_on_schedule()
        appointments = get_appointment_model_cls().objects.filter(
            subject_identifier=self.subject_identifier
        )
        self.assertEqual(appointments.count(), 4)
        # create report for baseline visit
        get_related_visit_model_cls().objects.create(
            appointment=appointments[0],
            report_datetime=appointments[0].appt_datetime,
            reason=SCHEDULED,
        )
        appointment = get_appointment_model_cls().objects.get(id=appointments[1].id)
        appointment.appt_timing = MISSED_APPT
        appointment.save()
        appointment.refresh_from_db()

        self.assertEqual(appointment.appt_status, IN_PROGRESS_APPT)

        # change subject visit to scheduled is ignored, forces
        # subject_visit.reason == MISSED_VISIT
        # you need to go to the appt
        self.assertEqual(appointment.appt_timing, MISSED_APPT)
        subject_visit = get_related_visit_model_cls().objects.get(
            appointment__visit_code=appointment.visit_code
        )
        subject_visit.reason = SCHEDULED
        self.assertRaises(RelatedVisitReasonError, subject_visit.save)

        subject_visit.refresh_from_db()
        self.assertEqual(subject_visit.reason, MISSED_VISIT)
        self.assertEqual(appointment.appt_timing, MISSED_APPT)

        # change appt to scheduled updates subject_visit
        appointment.appt_timing = ONTIME_APPT
        appointment.save()
        appointment.refresh_from_db()
        subject_visit.refresh_from_db()
        self.assertEqual(appointment.appt_timing, ONTIME_APPT)
        self.assertEqual(subject_visit.document_status, INCOMPLETE)
        self.assertEqual(subject_visit.reason, SCHEDULED)



# Generated by Django 4.2.5 on 2023-09-25 15:43

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_appointment", "0043_alter_appointment_appt_status_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="appointment",
            name="appt_reason",
            field=models.CharField(
                choices=[
                    ("scheduled", "Scheduled visit (study)"),
                    ("unscheduled", "Routine / Unscheduled (non-study)"),
                ],
                help_text="The reason for visit from the visit report will be validated against this response. Refer to the protocol documentation for the definition of a scheduled appointment.",
                max_length=25,
                verbose_name="Reason for appointment",
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="appt_reason",
            field=models.CharField(
                choices=[
                    ("scheduled", "Scheduled visit (study)"),
                    ("unscheduled", "Routine / Unscheduled (non-study)"),
                ],
                help_text="The reason for visit from the visit report will be validated against this response. Refer to the protocol documentation for the definition of a scheduled appointment.",
                max_length=25,
                verbose_name="Reason for appointment",
            ),
        ),
    ]



from __future__ import annotations

import uuid
from datetime import datetime
from typing import TYPE_CHECKING, Union
from uuid import UUID

from dateutil.relativedelta import relativedelta
from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from edc_document_status.model_mixins import DocumentStatusModelMixin
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_metadata.model_mixins import MetadataHelperModelMixin
from edc_offstudy.model_mixins import OffstudyNonCrfModelMixin
from edc_timepoint.model_mixins import TimepointModelMixin
from edc_utils import formatted_datetime, to_utc
from edc_visit_schedule.model_mixins import VisitScheduleModelMixin
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_schedule.subject_schedule import NotOnScheduleError
from edc_visit_schedule.utils import is_baseline

from ..constants import CANCELLED_APPT, IN_PROGRESS_APPT
from ..exceptions import AppointmentDatetimeError, UnknownVisitCode
from ..managers import AppointmentManager
from ..utils import raise_on_appt_may_not_be_missed, update_appt_status
from .appointment_fields_model_mixin import AppointmentFieldsModelMixin
from .appointment_methods_model_mixin import AppointmentMethodsModelMixin
from .missed_appointment_model_mixin import MissedAppointmentModelMixin
from .window_period_model_mixin import WindowPeriodModelMixin

if TYPE_CHECKING:
    from ..models import Appointment

# if django.VERSION[0] < 4:
constraints = []
unique_together = (
    (
        "subject_identifier",
        "visit_schedule_name",
        "schedule_name",
        "visit_code",
        "timepoint",
        "visit_code_sequence",
    ),
    ("subject_identifier", "visit_schedule_name", "schedule_name", "appt_datetime"),
)
# else:
#     constraints = [
#         UniqueConstraint(
#             "subject_identifier",
#             Lower("visit_schedule_name"),
#             Lower("schedule_name"),
#             Lower("visit_code"),
#             "timepoint",
#             "visit_code_sequence",
#             name="subject_id_visit_timepoint_seq_d98fxg_unique",
#         ),
#         UniqueConstraint(
#             "subject_identifier",
#             Lower("visit_schedule_name"),
#             Lower("schedule_name"),
#             "appt_datetime",
#             name="subject_id_visit_appt_datetime_9u3drr_unique",
#         ),
#     ]
#     unique_together = ()


class AppointmentModelMixin(
    NonUniqueSubjectIdentifierFieldMixin,
    AppointmentFieldsModelMixin,
    AppointmentMethodsModelMixin,
    TimepointModelMixin,
    MissedAppointmentModelMixin,
    WindowPeriodModelMixin,
    VisitScheduleModelMixin,
    DocumentStatusModelMixin,
    MetadataHelperModelMixin,
    OffstudyNonCrfModelMixin,
):

    """Mixin for the appointment model only.

    Only one appointment per subject visit+visit_code_sequence.

    Attribute 'visit_code_sequence' should be populated by the system.
    """

    metadata_helper_instance_attr = None

    offschedule_compare_dates_as_datetimes = False

    objects = AppointmentManager()

    def __str__(self) -> str:
        return f"{self.subject_identifier} {self.visit_code}.{self.visit_code_sequence}"

    def save(self: Appointment, *args, **kwargs):
        if not kwargs.get("update_fields", None):
            if self.id and is_baseline(instance=self):
                visit_schedule = site_visit_schedules.get_visit_schedule(
                    self.visit_schedule_name
                )
                schedule = visit_schedule.schedules.get(self.schedule_name)
                try:
                    onschedule_obj = django_apps.get_model(
                        schedule.onschedule_model
                    ).objects.get(
                        subject_identifier=self.subject_identifier,
                        onschedule_datetime__lte=to_utc(self.appt_datetime)
                        + relativedelta(seconds=1),
                    )
                except ObjectDoesNotExist as e:
                    dte_as_str = formatted_datetime(self.appt_datetime)
                    raise NotOnScheduleError(
                        "Subject is not on a schedule. Using subject_identifier="
                        f"`{self.subject_identifier}` and appt_datetime=`{dte_as_str}`."
                        f"Got {e}"
                    )
                if self.appt_datetime > onschedule_obj.onschedule_datetime:
                    # update appointment timepoints
                    schedule.put_on_schedule(
                        subject_identifier=self.subject_identifier,
                        onschedule_datetime=to_utc(self.appt_datetime),
                        skip_baseline=True,
                    )
            else:
                self.validate_appt_datetime_not_after_next()
            raise_on_appt_may_not_be_missed(appointment=self)
            self.update_subject_visit_reason_or_raise()
            if self.appt_status != IN_PROGRESS_APPT and getattr(
                settings, "EDC_APPOINTMENT_CHECK_APPT_STATUS", True
            ):
                update_appt_status(self)
        super().save(*args, **kwargs)

    def natural_key(self) -> tuple:
        return (
            self.subject_identifier,
            self.visit_schedule_name,
            self.schedule_name,
            self.visit_code,
            self.visit_code_sequence,
        )

    @property
    def str_pk(self: Appointment) -> Union[str, uuid.UUID]:
        if isinstance(self.id, UUID):
            return str(self.pk)
        return self.pk

    def validate_appt_datetime_not_after_next(self) -> None:
        if self.appt_status != CANCELLED_APPT and self.appt_datetime and self.relative_next:
            if self.appt_datetime >= self.relative_next.appt_datetime:
                appt_datetime = formatted_datetime(self.appt_datetime)
                next_appt_datetime = formatted_datetime(self.relative_next.appt_datetime)
                raise AppointmentDatetimeError(
                    "Datetime cannot be on or after next appointment datetime. "
                    f"Got {appt_datetime} >= {next_appt_datetime}. "
                    f"Perhaps catch this in the form. See {self}."
                )

    @property
    def title(self: Appointment) -> str:
        if not self.schedule.visits.get(self.visit_code):
            valid_visit_codes = [v for v in self.schedule.visits]
            raise UnknownVisitCode(
                "Unknown visit code specified for existing apointment instance. "
                "Has the appointments schedule changed? Expected one of "
                f"{valid_visit_codes}. Got {self.visit_code}. "
                f"See {self}."
            )
        title = self.schedule.visits.get(self.visit_code).title
        if self.visit_code_sequence > 0:
            title = f"{title}.{self.visit_code_sequence}"
        return title

    @property
    def report_datetime(self: Appointment) -> datetime:
        return self.appt_datetime

    class Meta:
        abstract = True
        constraints = constraints
        unique_together = unique_together
        ordering = ("timepoint", "visit_code_sequence")
        indexes = [
            models.Index(
                fields=[
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "timepoint",
                    "visit_code_sequence",
                ]
            )
        ]



# Generated by Django 3.2.4 on 2021-06-23 23:25

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_appointment", "0024_auto_20200911_0425"),
    ]

    operations = [
        migrations.AlterField(
            model_name="appointment",
            name="appt_reason",
            field=models.CharField(
                choices=[("scheduled", "Routine / Scheduled"), ("unscheduled", "Unscheduled")],
                help_text="The visit report's `reason for visit` will be validated against this. Refer to the protocol's documentation for the definition of a `scheduled` visit.",
                max_length=25,
                verbose_name="Reason for appointment",
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="appt_reason",
            field=models.CharField(
                choices=[("scheduled", "Routine / Scheduled"), ("unscheduled", "Unscheduled")],
                help_text="The visit report's `reason for visit` will be validated against this. Refer to the protocol's documentation for the definition of a `scheduled` visit.",
                max_length=25,
                verbose_name="Reason for appointment",
            ),
        ),
    ]



from __future__ import annotations

from django.utils.translation import gettext as _
from edc_model_admin.list_filters import FutureDateListFilter


class AppointmentListFilter(FutureDateListFilter):
    title = _("Appointment date")

    parameter_name = "appt_datetime"
    field_name = "appt_datetime"



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING
from zoneinfo import ZoneInfo

from django.db.models.deletion import ProtectedError
from edc_facility.exceptions import FacilityError
from edc_facility.utils import get_facility

from .appointment_creator import AppointmentCreator, CreateAppointmentError

if TYPE_CHECKING:
    from edc_visit_schedule.schedule import Schedule
    from edc_visit_schedule.visit_schedule import VisitSchedule

    from ..models import Appointment


class AppointmentsCreator:

    """Note: Appointments are created using this class by
    the visit schedule.

    See also: edc_visit_schedule SubjectSchedule

    """

    appointment_creator_cls = AppointmentCreator

    def __init__(
        self,
        subject_identifier: str | None = None,
        visit_schedule: VisitSchedule | None = None,
        schedule: Schedule | None = None,
        report_datetime: datetime | None = None,
        appointment_model: Appointment = None,
        skip_baseline: bool | None = None,
    ):
        self.subject_identifier = subject_identifier
        self.visit_schedule = visit_schedule
        self.schedule = schedule
        self.report_datetime = report_datetime
        self.appointment_model = appointment_model
        self.skip_baseline = skip_baseline

    def create_appointments(
        self, base_appt_datetime=None, taken_datetimes=None
    ) -> list[Appointment]:
        """Creates appointments when called by post_save signal.

        Timepoint datetimes are adjusted according to the available
        days in the facility.
        """
        appointments = []
        taken_datetimes = taken_datetimes or []
        base_appt_datetime = (base_appt_datetime or self.report_datetime).astimezone(
            ZoneInfo("UTC")
        )
        timepoint_dates = self.schedule.visits.timepoint_dates(dt=base_appt_datetime)
        for visit, timepoint_datetime in timepoint_dates.items():
            try:
                facility = get_facility(visit.facility_name)
            except FacilityError as e:
                raise CreateAppointmentError(
                    f"{e} See {repr(visit)}. Got facility_name={visit.facility_name}"
                )
            appointment = self.update_or_create_appointment(
                visit=visit,
                taken_datetimes=taken_datetimes,
                timepoint_datetime=timepoint_datetime,
                facility=facility,
            )
            appointments.append(appointment)
            taken_datetimes.append(appointment.appt_datetime)
        return appointments

    def update_or_create_appointment(self, **kwargs) -> Appointment:
        """Updates or creates an appointment for this subject
        for the visit.
        """
        appointment_creator = self.appointment_creator_cls(
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule.name,
            schedule_name=self.schedule.name,
            appointment_model=self.appointment_model,
            skip_baseline=self.skip_baseline,
            **kwargs,
        )
        return appointment_creator.appointment

    def delete_unused_appointments(self) -> None:
        appointments = self.appointment_model.objects.filter(
            subject_identifier=self.subject_identifier,
            visit_schedule_name=self.visit_schedule.name,
            schedule_name=self.schedule.name,
        )
        for appointment in appointments:
            try:
                appointment.delete()
            except ProtectedError:
                pass



from datetime import datetime
from typing import Any, Optional, Protocol, TypeVar, Union
from uuid import UUID

from django.db import models
from edc_visit_schedule.schedule import Schedule


class AppointmentModelStub(Protocol):
    id: Union[UUID, models.UUIDField]
    pk: Union[UUID, models.UUIDField]
    subject_identifier: Union[str, models.CharField]
    appt_datetime: Union[datetime, models.DateTimeField]
    visit_code: Union[str, models.CharField]
    visit_code_sequence: Union[int, models.IntegerField]
    visit_schedule_name: Union[str, models.CharField]
    schedule_name: Union[str, models.CharField]
    facility_name: Union[str, models.CharField]
    timepoint: Union[int, models.IntegerField]
    timepoint_datetime: datetime
    schedule: Schedule
    _meta: Any

    objects: models.Manager

    last_visit_code_sequence: Optional[int]
    next: "AppointmentModelStub"
    previous: "AppointmentModelStub"
    get_next: "AppointmentModelStub"

    def save(self, *args, **kwargs) -> None:
        ...

    def natural_key(self) -> tuple:
        ...

    def get_previous(self) -> "AppointmentModelStub":
        ...

    @classmethod
    def related_visit_model_attr(cls) -> str:
        ...


TAppointmentModelStub = TypeVar("TAppointmentModelStub", bound="AppointmentModelStub")



from dateutil.relativedelta import relativedelta
from edc_visit_schedule.constants import MONTH0, MONTH1, MONTH2, MONTH3, MONTH4
from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.visit import Crf, FormsCollection, Visit
from edc_visit_schedule.visit_schedule import VisitSchedule

app_label = "edc_appointment_app"


crfs = FormsCollection(
    Crf(show_order=1, model=f"{app_label}.nextappointmentcrf", required=True),
)


visit0 = Visit(
    code=MONTH0,
    title=MONTH0,
    timepoint=0,
    rbase=relativedelta(days=0),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs,
    facility_name="5-day-clinic",
)

visit1 = Visit(
    code=MONTH1,
    title=MONTH1,
    timepoint=1,
    rbase=relativedelta(months=1),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs,
    facility_name="5-day-clinic",
)

visit2 = Visit(
    code=MONTH2,
    title=MONTH2,
    timepoint=2,
    rbase=relativedelta(months=2),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs,
    facility_name="5-day-clinic",
)

visit3 = Visit(
    code=MONTH3,
    title=MONTH3,
    timepoint=3,
    rbase=relativedelta(months=3),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs,
    facility_name="5-day-clinic",
)

visit4 = Visit(
    code=MONTH4,
    title=MONTH4,
    timepoint=4,
    rbase=relativedelta(months=4),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs,
    facility_name="5-day-clinic",
)

schedule = Schedule(
    name="schedule",
    onschedule_model=f"{app_label}.onschedulesix",
    offschedule_model=f"{app_label}.offschedulesix",
    consent_model=f"{app_label}.subjectconsent",
    appointment_model="edc_appointment.appointment",
)

schedule.add_visit(visit0)
schedule.add_visit(visit1)
schedule.add_visit(visit2)
schedule.add_visit(visit3)
schedule.add_visit(visit4)

visit_schedule6 = VisitSchedule(
    name="visit_schedule6",
    offstudy_model=f"{app_label}.subjectoffstudy",
    death_report_model=f"{app_label}.deathreport",
)

visit_schedule6.add_schedule(schedule)



from __future__ import annotations

from django.db import models
from django.db.models import PROTECT
from django.utils.translation import gettext_lazy as _
from edc_facility.utils import get_health_facility_model
from edc_visit_schedule.models import VisitSchedule


class NextAppointmentCrfModelMixin(models.Model):
    health_facility = models.ForeignKey(
        get_health_facility_model(),
        on_delete=PROTECT,
        null=True,
        blank=True,
    )

    appt_date = models.DateField(
        verbose_name=_("Next scheduled routine/facility appointment"),
        null=True,
        blank=False,
        help_text=_("Should fall on an Integrated clinic day"),
    )

    info_source = models.ForeignKey(
        "edc_appointment.infosources",
        verbose_name=_("What is the source of this appointment date"),
        max_length=15,
        on_delete=PROTECT,
        null=True,
        blank=False,
    )

    # named this way to not conflict with property `visit_schedule`
    # see also edc-crf
    visitschedule = models.ForeignKey(
        VisitSchedule,
        on_delete=PROTECT,
        verbose_name=_("Which study visit code is closest to this appointment date"),
        max_length=15,
        null=True,
        blank=False,
        help_text=_(
            "Click SAVE to let the EDC suggest. Once selected, interim appointments will "
            "be flagged as not required/missed."
        ),
    )

    allow_create_interim = models.BooleanField(
        verbose_name=_("Override date check?"),
        default=False,
        help_text=(
            "Override date check to allow the EDC to create an interim appointment if the "
            "date is within window period of the current appointment."
        ),
    )

    class Meta:
        abstract = True
        verbose_name = "Next Appointment"
        verbose_name_plural = "Next Appointments"



from __future__ import annotations

import calendar
from typing import TYPE_CHECKING

from django.utils.translation import gettext_lazy as _
from edc_form_validators import INVALID_ERROR
from edc_utils.date import to_local

from ..utils import allow_clinic_on_weekend, get_allow_skipped_appt_using

if TYPE_CHECKING:
    from edc_facility.models import HealthFacility


class NextAppointmentCrfFormValidatorMixin:
    def __init__(self, **kwargs):
        self._clinic_days = None
        self._health_facility = None
        self.day_abbr = calendar.weekheader(3).split(" ")
        super().__init__(**kwargs)

    @property
    def visit_code_fld(self):
        return get_allow_skipped_appt_using().get(self._meta.model._meta.label_lower)[1]

    @property
    def dt_fld(self):
        return get_allow_skipped_appt_using().get(self._meta.model._meta.label_lower)[1]

    @property
    def clinic_days(self) -> list[int]:
        if not self._clinic_days:
            if self.cleaned_data.get("health_facility"):
                self._clinic_days = self.health_facility.clinic_days
        return self._clinic_days

    def validate_date_is_on_clinic_day(self):
        if appt_date := self.cleaned_data.get("appt_date"):
            if appt_date <= to_local(self.cleaned_data.get("report_datetime")).date():
                raise self.raise_validation_error(
                    {"appt_date": "Cannot be on or before the report datetime"}, INVALID_ERROR
                )
            if not allow_clinic_on_weekend() and appt_date.weekday() > calendar.FRIDAY:
                raise self.raise_validation_error(
                    {
                        "appt_date": _("Expected %(mon)s-%(fri)s. Got %(day)s")
                        % {
                            "mon": self.day_abbr[calendar.MONDAY],
                            "fri": self.day_abbr[calendar.FRIDAY],
                            "day": self.day_abbr[appt_date.weekday()],
                        }
                    },
                    INVALID_ERROR,
                )
            elif self.cleaned_data.get("subject_visit").site:
                if self.clinic_days and appt_date.weekday() not in self.clinic_days:
                    raise self.raise_validation_error(
                        {
                            "appt_date": _(
                                "Invalid clinic day. Expected %(expected)s. Got %(day_abbr)s"
                            )
                            % {
                                "expected": ", ".join(
                                    dict(zip(self.clinic_days, self.day_abbr)).values()
                                ),
                                "day_abbr": appt_date.strftime("%A"),
                            }
                        },
                        INVALID_ERROR,
                    )

    @property
    def health_facility(self) -> HealthFacility | None:
        if not self._health_facility:
            if self.cleaned_data.get("health_facility"):
                self._health_facility = self.cleaned_data.get("health_facility")
            else:
                raise self.raise_validation_error(
                    {"health_facility": _("This field is required.")}, INVALID_ERROR
                )
        return self._health_facility



import sys

from django.apps import AppConfig as DjangoAppConfig
from django.core.checks.registry import register

from .system_checks import context_processors_check


class AppConfig(DjangoAppConfig):
    _holidays: dict = {}
    name = "edc_appointment"
    verbose_name = "Appointments"
    has_exportable_data = True
    include_in_administration_section = True

    def ready(self):
        sys.stdout.write(f"Loading {self.verbose_name} ...\n")
        register(context_processors_check)
        sys.stdout.write(f" Done loading {self.verbose_name}.\n")



import datetime as dt
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import relativedelta
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase, override_settings
from edc_consent import site_consents
from edc_constants.constants import NOT_APPLICABLE, PATIENT
from edc_facility import import_holidays
from edc_facility.models import HealthFacility, HealthFacilityTypes
from edc_utils import get_utcnow
from edc_visit_schedule.apps import populate_visit_schedule
from edc_visit_schedule.models import VisitSchedule
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED
from edc_visit_tracking.utils import get_related_visit_model_cls

from edc_appointment.constants import SKIPPED_APPT
from edc_appointment.models import Appointment, InfoSources
from edc_appointment_app.consents import v1_consent
from edc_appointment_app.forms import NextAppointmentCrfForm
from edc_appointment_app.models import NextAppointmentCrf, SubjectConsent
from edc_appointment_app.visit_schedule import visit_schedule6

utc = ZoneInfo("UTC")
tz = ZoneInfo("Africa/Dar_es_Salaam")


class TestNextAppointmentCrf(TestCase):
    @time_machine.travel(dt.datetime(2019, 6, 11, 8, 00, tzinfo=utc))
    def setUp(self):
        import_holidays()
        self.user = User.objects.create_superuser("user_login", "u@example.com", "pass")

        site_visit_schedules._registry = {}
        site_visit_schedules.loaded = False
        site_visit_schedules.register(visit_schedule6)
        site_consents.registry = {}
        site_consents.register(v1_consent)

        populate_visit_schedule()

        self.subject_identifier = "101-40990029-4"
        identity = "123456789"
        subject_consent = SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier,
            consent_datetime=get_utcnow() - relativedelta(days=10),
            identity=identity,
            confirm_identity=identity,
            dob=get_utcnow() - relativedelta(years=25),
        )

        # put subject on schedule
        _, schedule = site_visit_schedules.get_by_onschedule_model(
            "edc_appointment_app.onschedulesix"
        )
        schedule.put_on_schedule(
            subject_identifier=subject_consent.subject_identifier,
            onschedule_datetime=subject_consent.consent_datetime,
        )

    @override_settings(
        EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING={
            "edc_appointment_app.nextappointmentcrf": ("appt_date", "visitschedule")
        }
    )
    @time_machine.travel(dt.datetime(2019, 6, 11, 8, 00, tzinfo=utc))
    def test_ok(self):
        self.assertEqual(5, Appointment.objects.all().count())
        appointment = Appointment.objects.get(timepoint=0)
        subject_visit_model_cls = get_related_visit_model_cls()
        subject_visit_model_cls.objects.create(
            report_datetime=appointment.appt_datetime,
            appointment=appointment,
            reason=SCHEDULED,
        )

    @override_settings(
        EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING={
            "edc_appointment_app.nextappointmentcrf": ("appt_date", "visitschedule")
        }
    )
    @time_machine.travel(dt.datetime(2019, 6, 11, 8, 00, tzinfo=utc))
    def test_next_appt_ok(self):
        self.assertEqual(5, Appointment.objects.all().count())
        appointment = Appointment.objects.get(timepoint=0)
        subject_visit = get_related_visit_model_cls().objects.create(
            appointment=appointment,
            reason=SCHEDULED,
            report_datetime=appointment.report_datetime,
        )
        data = dict(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            appt_date=(appointment.appt_datetime + relativedelta(months=3)).date(),
            visitschedule=VisitSchedule.objects.get(visit_code="1000"),
            info_source=InfoSources.objects.get(name=PATIENT),
        )
        form = NextAppointmentCrfForm(data=data)
        form.is_valid()

        self.assertIn("visitschedule", form._errors)
        self.assertIn("1030", str(form._errors.get("visitschedule")))

        data.update(visitschedule=VisitSchedule.objects.get(visit_code="1010"))
        form = NextAppointmentCrfForm(data=data)
        form.is_valid()
        self.assertIn("1030", str(form._errors.get("visitschedule")))

        data.update(visitschedule=VisitSchedule.objects.get(visit_code="1020"))
        form = NextAppointmentCrfForm(data=data)
        form.is_valid()
        self.assertIn("1030", str(form._errors.get("visitschedule")))

        data.update(visitschedule=VisitSchedule.objects.get(visit_code="1030"))
        form = NextAppointmentCrfForm(data=data)
        form.is_valid()
        self.assertNotIn("1030", str(form._errors.get("visitschedule")))

        # assert NextAppointmentCrf exists
        form.save()
        try:
            NextAppointmentCrf.objects.get(subject_visit=subject_visit)
        except ObjectDoesNotExist:
            self.fail("NextAppointmentCrf unexpectedly does not exist")

        # assert skipped over 1010, 1020
        apppointment = Appointment.objects.get(visit_code="1010")
        self.assertEqual(apppointment.appt_status, SKIPPED_APPT)
        self.assertEqual(apppointment.appt_timing, NOT_APPLICABLE)
        apppointment = Appointment.objects.get(visit_code="1020")
        self.assertEqual(apppointment.appt_status, SKIPPED_APPT)
        self.assertEqual(apppointment.appt_timing, NOT_APPLICABLE)

    @override_settings(
        EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING={
            "edc_appointment_app.nextappointmentcrf": ("appt_date", "visitschedule")
        }
    )
    @time_machine.travel(dt.datetime(2019, 6, 11, 8, 00, tzinfo=utc))
    def test_next_appt_with_health_facility(self):
        self.assertEqual(get_utcnow().weekday(), 1)  # tues
        health_facility_type = HealthFacilityTypes.objects.create(
            name="Integrated", display_name="Integrated"
        )
        health_facility = HealthFacility.objects.create(
            name="integrated_facility",
            health_facility_type=health_facility_type,
            mon=False,
            tue=True,
            wed=False,
            thu=True,
            fri=False,
            sat=False,
            sun=False,
        )
        self.assertEqual(5, Appointment.objects.all().count())
        appointment = Appointment.objects.get(timepoint=0)
        subject_visit_model_cls = get_related_visit_model_cls()
        subject_visit = subject_visit_model_cls.objects.create(
            appointment=appointment,
            reason=SCHEDULED,
            report_datetime=appointment.report_datetime,
        )

        data = dict(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            appt_date=(
                appointment.appt_datetime + relativedelta(months=3) + relativedelta(days=3)
            ).date(),
            info_source=InfoSources.objects.get(name=PATIENT),
            visitschedule=VisitSchedule.objects.get(visit_code="1030"),
            health_facility=health_facility.id,
        )
        self.assertEqual(data.get("appt_date").weekday(), 4)
        form = NextAppointmentCrfForm(data=data)
        form.is_valid()
        self.assertIn("appt_date", form._errors)
        self.assertIn("Invalid clinic day", str(form._errors.get("appt_date")))

        data.update(appt_date=data.get("appt_date") + relativedelta(days=1))
        form = NextAppointmentCrfForm(data=data)
        form.is_valid()
        self.assertIn("appt_date", form._errors)
        self.assertIn("Expected Mon-Fri", str(form._errors.get("appt_date")))

        data.update(appt_date=data.get("appt_date") - relativedelta(days=2))
        form = NextAppointmentCrfForm(data=data)
        form.is_valid()
        self.assertNotIn("appt_date", form._errors)

        form = NextAppointmentCrfForm(data=data)
        form.is_valid()
        self.assertEqual({}, form._errors)

    @override_settings(
        EDC_APPOINTMENT_ALLOW_SKIPPED_APPT_USING={
            "edc_appointment_app.nextappointmentcrf": ("appt_date", "visitschedule")
        },
        LANGUAGE_CODE="sw",
    )
    @time_machine.travel(dt.datetime(2019, 6, 11, 8, 00, tzinfo=tz))
    def test_next_appt_with_health_facility_tz(self):
        self.assertEqual(get_utcnow().weekday(), 1)  # tues
        health_facility_type = HealthFacilityTypes.objects.create(
            name="Integrated", display_name="Integrated"
        )
        health_facility = HealthFacility.objects.create(
            name="integrated_facility",
            health_facility_type=health_facility_type,
            mon=False,
            tue=True,
            wed=False,
            thu=True,
            fri=False,
            sat=False,
            sun=False,
        )
        appointment = Appointment.objects.get(timepoint=0)
        subject_visit_model_cls = get_related_visit_model_cls()
        subject_visit = subject_visit_model_cls.objects.create(
            appointment=appointment,
            reason=SCHEDULED,
            report_datetime=appointment.report_datetime,
        )
        data = dict(
            subject_visit=subject_visit,
            report_datetime=subject_visit.report_datetime,
            appt_date=(
                appointment.appt_datetime + relativedelta(months=3) + relativedelta(days=2)
            ).date(),
            info_source=InfoSources.objects.get(name=PATIENT),
            visitschedule=VisitSchedule.objects.get(visit_code="1030"),
            health_facility=health_facility.id,
        )
        self.assertEqual(data.get("appt_date").weekday(), 3)



# Generated by Django 4.1.2 on 2023-01-24 15:14

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("edc_appointment", "0033_alter_appointment_appt_reason_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="AppointmentType",
            fields=[
                (
                    "name",
                    models.CharField(
                        db_index=True,
                        help_text="This is the stored value, required",
                        max_length=250,
                        unique=True,
                        verbose_name="Stored value",
                    ),
                ),
                (
                    "plural_name",
                    models.CharField(max_length=250, null=True, verbose_name="Plural name"),
                ),
                (
                    "display_name",
                    models.CharField(
                        db_index=True,
                        help_text="(suggest 40 characters max.)",
                        max_length=250,
                        unique=True,
                        verbose_name="Name",
                    ),
                ),
                (
                    "display_index",
                    models.IntegerField(
                        db_index=True,
                        default=0,
                        help_text="Index to control display order if not alphabetical, not required",
                        verbose_name="display index",
                    ),
                ),
                (
                    "field_name",
                    models.CharField(
                        blank=True,
                        editable=False,
                        help_text="Not required",
                        max_length=25,
                        null=True,
                    ),
                ),
                ("version", models.CharField(default="1.0", editable=False, max_length=35)),
                ("id", models.AutoField(primary_key=True, serialize=False)),
            ],
            options={
                "verbose_name": "Appointment Type",
                "verbose_name_plural": "Appointment Types",
            },
        ),
        migrations.AlterField(
            model_name="appointment",
            name="appt_type",
            field=models.CharField(
                choices=[
                    ("clinic", "In clinic"),
                    ("home", "At home"),
                    ("hospital", "In hospital"),
                    ("telephone", "By telephone"),
                ],
                default="clinic",
                max_length=20,
                verbose_name="Appointment type",
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="appt_type",
            field=models.CharField(
                choices=[
                    ("clinic", "In clinic"),
                    ("home", "At home"),
                    ("hospital", "In hospital"),
                    ("telephone", "By telephone"),
                ],
                default="clinic",
                max_length=20,
                verbose_name="Appointment type",
            ),
        ),
        migrations.AddField(
            model_name="appointment",
            name="appt_type_new",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_appointment.appointmenttype",
                verbose_name="Appointment type",
            ),
        ),
        migrations.AddField(
            model_name="historicalappointment",
            name="appt_type_new",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_appointment.appointmenttype",
                verbose_name="Appointment type",
            ),
        ),
    ]
